# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pandas as pd
import numpy as np
import copy
import warnings
from packaging import version
from chainladder.core.base import TriangleBase
from chainladder.utils.sparse import sp
from chainladder.core.slice import VirtualColumns
from chainladder.core.correlation import DevelopmentCorrelation, ValuationCorrelation
from chainladder.utils.utility_functions import concat, num_to_nan, num_to_value
from chainladder import options

try:
    import dask.bag as db
except:
    db = None


class Triangle(TriangleBase):
    """
    The core data structure of the chainladder package

    Parameters
    ----------
    data: DataFrame
        A single dataframe that contains columns represeting all other
        arguments to the Triangle constructor
    origin: str or list
         A representation of the accident, reporting or more generally the
         origin period of the triangle that will map to the Origin dimension
    development: str or list
        A representation of the development/valuation periods of the triangle
        that will map to the Development dimension
    columns: str or list
        A representation of the numeric data of the triangle that will map to
        the columns dimension.  If None, then a single 'Total' key will be
        generated.
    index: str or list or None
        A representation of the index of the triangle that will map to the
        index dimension.  If None, then a single 'Total' key will be generated.
    origin_format: optional str
        A string representation of the date format of the origin arg. If
        omitted then date format will be inferred by pandas.
    development_format: optional str
        A string representation of the date format of the development arg. If
        omitted then date format will be inferred by pandas.
    cumulative: bool
        Whether the triangle is cumulative or incremental.  This attribute is
        required to use the ``grain`` and ``dev_to_val`` methods and will be
        automatically set when invoking ``cum_to_incr`` or ``incr_to_cum`` methods.
    trailing: bool
        When partial origin periods are present, setting trailing to True will
        ensure the most recent origin period is a full period and the oldest
        origin is partial. If full origin periods are present in the data, then
        trailing has no effect.

    Attributes
    ----------
    index: Series
        Represents all available levels of the index dimension.
    columns: Series
        Represents all available levels of the value dimension.
    origin: DatetimeIndex
        Represents all available levels of the origin dimension.
    development: Series
        Represents all available levels of the development dimension.
    key_labels: list
        Represents the ``index`` axis labels
    virtual_columns: Series
        Represents the subset of columns of the triangle that are virtual.
    valuation: DatetimeIndex
        Represents all valuation dates of each cell in the Triangle.
    origin_grain: str
        The grain of the origin vector ('Y', 'S', 'Q', 'M')
    development_grain: str
        The grain of the development vector ('Y', 'S', 'Q', 'M')
    shape: tuple
        The 4D shape of the triangle instance with axes corresponding to (index, columns, origin, development)
    link_ratio, age_to_age
        Displays age-to-age ratios for the triangle.
    valuation_date : date
        The latest valuation date of the data
    loc: Triangle
        pandas-style ``loc`` accessor
    iloc: Triangle
        pandas-style ``iloc`` accessor
    latest_diagonal: Triangle
        The latest diagonal of the triangle
    is_cumulative: bool
        Whether the triangle is cumulative or not
    is_ultimate: bool
        Whether the triangle has an ultimate valuation
    is_full: bool
        Whether lower half of Triangle has been filled in
    is_val_tri:
        Whether the triangle development period is expressed as valuation
        periods.
    values: array
        4D numpy array underlying the Triangle instance
    T: Triangle
        Transpose index and columns of object.  Only available when Triangle is
        convertible to DataFrame.
    """

    def __init__(
        self,
        data=None,
        origin=None,
        development=None,
        columns=None,
        index=None,
        origin_format=None,
        development_format=None,
        cumulative=None,
        array_backend=None,
        pattern=False,
        trailing=True,
        *args,
        **kwargs
    ):
        if data is None:
            return
        elif not isinstance(data, pd.DataFrame) and hasattr(data, "__dataframe__"):
            data = self._interchange_dataframe(data)
        index, columns, origin, development = self._input_validation(
            data, index, columns, origin, development
        )

        self.columns_label = columns
        self.origin_label = origin

        # Handle any ultimate vectors in triangles separately
        data, ult = self._split_ult(data, index, columns, origin, development)
        # Conform origins and developments to datetimes and determine lowest grains
        origin_date = self._to_datetime(data, origin, format=origin_format).rename(
            "__origin__"
        )
        self.origin_grain = self._get_grain(
            origin_date, trailing=trailing, kind="origin"
        )

        development_date = self._set_development(
            data, development, development_format, origin_date
        )

        self.development_grain = self._get_grain(
            development_date, trailing=trailing, kind="development"
        )

        origin_date = origin_date.dt.to_period(self.origin_grain).dt.to_timestamp(
            how="s"
        )

        development_date = development_date.dt.to_period(
            self.development_grain
        ).dt.to_timestamp(how="e")

        # Aggregate dates to the origin/development grains
        data_agg = self._aggregate_data(
            data, origin_date, development_date, index, columns
        )

        # Fill in missing periods with zeros
        date_axes = self._get_date_axes(
            data_agg["__origin__"],
            data_agg["__development__"],
            self.origin_grain,
            self.development_grain,
        )

        # Deal with labels
        if not index:
            index = ["Total"]
            self.index_label = index
            data_agg[index[0]] = "Total"

        self.kdims, key_idx = self._set_kdims(data_agg, index)
        self.vdims = np.array(columns)
        self.odims, orig_idx = self._set_odims(data_agg, date_axes)
        self.ddims, dev_idx = self._set_ddims(data_agg, date_axes)

        # Set remaining triangle properties
        val_date = data_agg["__development__"].max()
        val_date = val_date.compute() if hasattr(val_date, "compute") else val_date
        self.key_labels = index
        self.valuation_date = val_date

        if cumulative is None:
            warnings.warn(
                """
            The cumulative property of your triangle is not set. This may result in
            undesirable behavior. In a future release this will result in an error."""
            )

        self.is_cumulative = cumulative
        self.virtual_columns = VirtualColumns(self)
        self.is_pattern = pattern

        split = self.origin_grain.split("-")
        self.origin_grain = {"A": "Y", "2Q": "S"}.get(split[0], split[0])

        if len(split) == 1:
            self.origin_close = "DEC"
        else:
            self.origin_close = split[1]

        split = self.development_grain.split("-")
        self.development_grain = {"A": "Y", "2Q": "S"}.get(split[0], split[0])
        grain_sort = ["Y", "S", "Q", "M"]
        self.development_grain = grain_sort[
            max(
                grain_sort.index(self.origin_grain),
                grain_sort.index(self.development_grain),
            )
        ]

        # Coerce malformed triangles to something more predictible
        check_origin = (
            pd.period_range(
                start=self.odims.min(),
                end=self.valuation_date,
                freq=self.origin_grain.replace("S", "2Q"),
            )
            .to_timestamp()
            .values
        )

        if (
            len(check_origin) != len(self.odims)
            and pd.to_datetime(options.ULT_VAL) != self.valuation_date
            and not self.is_pattern
        ):
            self.odims = check_origin

        # Set the Triangle values
        coords, amts = self._set_values(data_agg, key_idx, columns, orig_idx, dev_idx)

        self.values = num_to_nan(
            sp(
                coords,
                amts,
                prune=True,
                has_duplicates=False,
                sorted=True,
                shape=(
                    len(self.kdims),
                    len(self.vdims),
                    len(self.odims),
                    len(self.ddims),
                ),
            )
        )
        # Deal with array backend
        self.array_backend = "sparse"
        if array_backend is None:
            array_backend = options.ARRAY_BACKEND
        if not options.AUTO_SPARSE or array_backend == "cupy":
            self.set_backend(array_backend, inplace=True)
        else:
            self = self._auto_sparse()
        self._set_slicers()
        # Deal with special properties
        if self.is_pattern:
            obj = self.dropna()
            self.odims = obj.odims
            self.ddims = obj.ddims
            self.values = obj.values
        if ult:
            obj = concat((self.dev_to_val().iloc[..., : len(ult.odims), :], ult), -1)
            obj = obj.val_to_dev()
            self.odims = obj.odims
            self.ddims = obj.ddims
            self.values = obj.values
            self.valuation_date = pd.Timestamp(options.ULT_VAL)

    @staticmethod
    def _split_ult(data, index, columns, origin, development):
        """Deal with triangles with ultimate values"""
        ult = None
        if (
            development
            and len(development) == 1
            and data[development[0]].dtype == "<M8[ns]"
        ):
            u = data[data[development[0]] == options.ULT_VAL].copy()
            if len(u) > 0 and len(u) != len(data):
                ult = Triangle(
                    u,
                    origin=origin,
                    development=development,
                    columns=columns,
                    index=index,
                )
                ult.ddims = pd.DatetimeIndex([options.ULT_VAL])
                data = data[data[development[0]] != options.ULT_VAL]
        return data, ult

    @property
    def index(self):
        return pd.DataFrame(list(self.kdims), columns=self.key_labels)

    @index.setter
    def index(self, value):
        self._len_check(self.index, value)
        if type(value) is pd.DataFrame:
            self.kdims = value.values
            self.key_labels = list(value.columns)
            self._set_slicers()
        else:
            raise TypeError("index must be a pandas DataFrame")

    @property
    def columns(self):
        return pd.Index(self.vdims, name="columns")

    @columns.setter
    def columns(self, value):
        self._len_check(self.columns, value)
        self.vdims = [value] if type(value) is str else value
        if type(self.vdims) is list:
            self.vdims = np.array(self.vdims)
        self._set_slicers()

    @property
    def origin(self):
        if self.is_pattern and len(self.odims) == 1:
            return pd.Series(["(All)"])
        else:
            freq = {
                "Y": (
                    "Y"
                    if version.Version(pd.__version__) >= version.Version("2.2.0")
                    else "A"
                ),
                "S": "2Q",
                "H": "2Q",
            }.get(self.origin_grain, self.origin_grain)
            freq = freq if freq == "M" else freq + "-" + self.origin_close
            return pd.DatetimeIndex(self.odims, name="origin").to_period(freq=freq)

    @origin.setter
    def origin(self, value):
        self._len_check(self.origin, value)
        freq = {
            "Y": "A" if float(".".join(pd.__version__.split(".")[:-1])) < 2.2 else "Y",
            "S": "2Q",
        }.get(self.origin_grain, self.origin_grain)
        freq = freq if freq == "M" else freq + "-" + self.origin_close
        value = pd.PeriodIndex(list(value), freq=freq)
        self.odims = value.to_timestamp().values

    @property
    def development(self):
        ddims = self.ddims.copy()
        if self.is_val_tri:
            formats = {"Y": "%Y", "S": "%YQ%q", "Q": "%YQ%q", "M": "%Y-%m"}
            ddims = ddims.to_period(freq=self.development_grain).strftime(
                formats[self.development_grain]
            )
        elif self.is_pattern:
            offset = self._dstep()["M"][self.development_grain]
            if self.is_ultimate:
                ddims[-1] = ddims[-2] + offset
            if self.is_cumulative:
                ddims = ["{}-Ult".format(ddims[i]) for i in range(len(ddims))]
            else:
                ddims = [
                    "{}-{}".format(ddims[i], ddims[i] + offset)
                    for i in range(len(ddims))
                ]
        return pd.Series(list(ddims), name="development")

    @development.setter
    def development(self, value):
        self._len_check(self.development, value)
        self.ddims = np.array([value] if type(value) is str else value)

    def set_index(self, value, inplace=False):
        """Sets the index of the Triangle"""
        if inplace:
            self.index = value
            return self
        else:
            new_obj = self.copy()
            return new_obj.set_index(value=value, inplace=True)

    @property
    def is_val_tri(self):
        return type(self.ddims) == pd.DatetimeIndex

    @property
    def is_full(self):
        return self.nan_triangle.sum().sum() == np.prod(self.shape[-2:])

    @property
    def is_ultimate(self):
        return sum(self.valuation >= options.ULT_VAL[:4]) > 0

    @property
    def latest_diagonal(self):
        return self[self.valuation == self.valuation_date].sum("development")

    @property
    def link_ratio(self):
        if not self.is_pattern:
            obj = (1 / self.iloc[..., :-1]) * self.iloc[..., 1:].values
            if not obj.is_full:
                obj = obj[obj.valuation < obj.valuation_date]
            if hasattr(obj, "w_"):
                w_ = obj.w_[..., : len(obj.odims), :]
                obj = obj * w_ if obj.shape == w_.shape else obj
            obj.is_pattern = True
            obj.is_cumulative = False
            obj.values = num_to_nan(obj.values)
            return obj
        else:
            return self

    @property
    def age_to_age(self):
        return self.link_ratio

    def incr_to_cum(self, inplace=False):
        """Method to convert an incremental triangle into a cumulative triangle.

        Parameters
        ----------
        inplace: bool
            Set to True will update the instance data attribute inplace

        Returns
        -------
            Updated instance of triangle accumulated along the origin
        """
        if inplace:
            xp = self.get_array_module()
            if not self.is_cumulative:
                if self.is_pattern:
                    if hasattr(self, "is_additive"):
                        if self.is_additive:
                            values = xp.nan_to_num(self.values[..., ::-1])
                            values = num_to_value(values, 0)
                            self.values = (
                                xp.cumsum(values, -1)[..., ::-1] * self.nan_triangle
                            )
                    else:
                        values = xp.nan_to_num(self.values[..., ::-1])
                        values = num_to_value(values, 1)
                        values = xp.cumprod(values, -1)[..., ::-1]
                        self.values = values * self.nan_triangle
                        values = num_to_value(values, self.get_array_module(values).nan)
                else:
                    if self.array_backend not in ["sparse", "dask"]:
                        self.values = (
                            xp.cumsum(xp.nan_to_num(self.values), 3)
                            * self.nan_triangle[None, None, ...]
                        )
                    else:
                        values = xp.nan_to_num(self.values)
                        nan_triangle = xp.nan_to_num(self.nan_triangle)
                        l1 = lambda i: values[..., 0 : i + 1]
                        l2 = lambda i: l1(i) * nan_triangle[..., i : i + 1]
                        l3 = lambda i: l2(i).sum(3, keepdims=True)
                        if db:
                            bag = db.from_sequence(range(self.shape[-1]))
                            bag = bag.map(l3)
                            out = bag.compute(scheduler="threads")
                        else:
                            out = [l3(i) for i in range(self.shape[-1])]
                        self.values = xp.concatenate(out, axis=3)
                    self.values = num_to_nan(self.values)
                self.is_cumulative = True
            return self
        else:
            new_obj = self.copy()
            return new_obj.incr_to_cum(inplace=True)

    def cum_to_incr(self, inplace=False):
        """Method to convert an cumlative triangle into a incremental triangle.

        Parameters
        ----------
            inplace: bool
                Set to True will update the instance data attribute inplace

        Returns
        -------
            Updated instance of triangle accumulated along the origin
        """
        if inplace:
            v = self.valuation_date
            if self.is_cumulative or self.is_cumulative is None:
                if self.is_pattern:
                    xp = self.get_array_module()
                    self.values = xp.nan_to_num(self.values)
                    values = num_to_value(self.values, 1)
                    diff = self.iloc[..., :-1] / self.iloc[..., 1:].values
                    self = concat(
                        (
                            diff,
                            self.iloc[..., -1],
                        ),
                        axis=3,
                    )
                    self.values = self.values * self.nan_triangle
                else:
                    diff = self.iloc[..., 1:] - self.iloc[..., :-1].values
                    self = concat((self.iloc[..., 0], diff), axis=3)
                self.is_cumulative = False
            self.valuation_date = v
            return self
        else:
            new_obj = self.copy()
            return new_obj.cum_to_incr(inplace=True)

    def _dstep(self):
        return {
            "M": {"Y": 12, "S": 6, "Q": 3, "M": 1},
            "Q": {"Y": 4, "S": 2, "Q": 1},
            "S": {"Y": 2, "S": 1},
            "Y": {"Y": 1},
        }

    def _val_dev(self, sign, inplace=False):
        backend = self.array_backend
        obj = self.set_backend("sparse")
        if not inplace:
            obj.values = obj.values.copy()
        scale = self._dstep()[obj.development_grain][obj.origin_grain]
        offset = np.arange(obj.shape[-2]) * scale
        min_slide = -offset.max()
        if (obj.values.coords[-2] == np.arange(1)).all():
            # Unique edge case #239
            offset = offset[-1:] * sign
        offset = offset[obj.values.coords[-2]] * sign  # [0]
        obj.values.coords[-1] = obj.values.coords[-1] + offset
        ddims = obj.valuation[obj.valuation <= obj.valuation_date]
        ddims = len(ddims.drop_duplicates())
        if ddims == 1 and sign == -1:
            ddims = len(obj.odims)
        if obj.values.density > 0 and obj.values.coords[-1].min() < 0:
            obj.values.coords[-1] = obj.values.coords[-1] - min(
                obj.values.coords[-1].min(), min_slide
            )
            ddims = np.max([np.max(obj.values.coords[-1]) + 1, ddims])
        obj.values.shape = tuple(list(obj.shape[:-1]) + [ddims])
        if options.AUTO_SPARSE == False or backend == "cupy":
            obj = obj.set_backend(backend)
        else:
            obj = obj._auto_sparse()
        return obj

    def dev_to_val(self, inplace=False):
        """Converts triangle from a development lag triangle to a valuation
        triangle.

        Parameters
        ----------
        inplace : bool
            Whether to mutate the existing Triangle instance or return a new
            one.

        Returns
        -------
        Triangle
            Updated instance of the triangle with valuation periods.

        """
        if self.is_val_tri:
            if inplace:
                return self
            else:
                return self.copy()
        is_cumulative = self.is_cumulative
        if self.is_full:
            if is_cumulative:
                obj = self.cum_to_incr(inplace=inplace)
            else:
                obj = self.copy()
            if self.is_ultimate:
                ultimate = obj.iloc[..., -1:]
                obj = obj.iloc[..., :-1]
        else:
            obj = self
        obj = obj._val_dev(1, inplace)
        ddims = obj.valuation[obj.valuation <= obj.valuation_date]
        obj.ddims = ddims.drop_duplicates().sort_values()
        if self.is_full:
            if self.is_ultimate:
                ultimate.ddims = pd.DatetimeIndex(ultimate.valuation[0:1])
                obj = concat((obj, ultimate), -1)
            if is_cumulative:
                obj = obj.incr_to_cum(inplace=inplace)
        return obj

    def val_to_dev(self, inplace=False):
        """Converts triangle from a valuation triangle to a development lag
        triangle.

        Parameters
        ----------
        inplace : bool
            Whether to mutate the existing Triangle instance or return a new
            one.

        Returns
        -------
            Updated instance of triangle with development lags
        """
        if not self.is_val_tri:
            if inplace:
                return self
            else:
                return self.copy()
        if self.is_ultimate and self.shape[-1] > 1:
            ultimate = self.iloc[..., -1:]
            ultimate.ddims = np.array([9999])
            obj = self.iloc[..., :-1]._val_dev(-1, inplace)
        else:
            obj = self.copy()._val_dev(-1, inplace)
        val_0 = obj.valuation[0]
        if self.ddims.shape[-1] == 1 and self.ddims[0] == self.valuation_date:
            origin_0 = pd.to_datetime(obj.odims[-1])
        else:
            origin_0 = pd.to_datetime(obj.odims[0])
        lag_0 = (val_0.year - origin_0.year) * 12 + val_0.month - origin_0.month + 1
        scale = self._dstep()["M"][obj.development_grain]
        obj.ddims = np.arange(obj.values.shape[-1]) * scale + lag_0
        prune = obj[obj.origin == obj.origin.max()]
        if self.is_ultimate and self.shape[-1] > 1:
            obj = obj.iloc[..., : (prune.valuation <= prune.valuation_date).sum()]
            obj = concat((obj, ultimate), -1)
        return obj

    def grain(self, grain="", trailing=False, inplace=False):
        """Changes the grain of a cumulative triangle.

        Parameters
        ----------
        grain : str
            The grain to which you want your triangle converted, specified as
            'OXDY' where X and Y can take on values of ``['Y', 'S', 'Q', 'M'
            ]`` For example, 'OYDY' for Origin Year/Development Year, 'OQDM'
            for Origin quarter/Development Month, etc.
        trailing : bool
            For partial origin years/quarters, trailing will set the year/quarter
            end to that of the latest available from the origin data.
        inplace : bool
            Whether to mutate the existing Triangle instance or return a new
            one.

        Returns
        -------
            Triangle
        """
        ograin_old, ograin_new = self.origin_grain, grain[1:2]
        dgrain_old, dgrain_new = self.development_grain, grain[-1]
        ograin_new = "S" if ograin_new == "H" else ograin_new
        valid = {
            "Y": ["Y"],
            "Q": ["Q", "S", "Y"],
            "M": ["Y", "S", "Q", "M"],
            "S": ["S", "Y"],
        }

        if ograin_new not in valid.get(ograin_old, []) or dgrain_new not in valid.get(
            dgrain_old, []
        ):
            raise ValueError("New grain not compatible with existing grain")

        if (
            self.is_cumulative is None
            and dgrain_old != dgrain_new
            and self.shape[-1] > 1
        ):
            raise AttributeError(
                "The is_cumulative attribute must be set before using grain method."
            )

        if valid["M"].index(ograin_new) > valid["M"].index(dgrain_new):
            raise ValueError("Origin grain must be coarser than development grain")

        if self.is_full and not self.is_ultimate and not self.is_val_tri:
            warnings.warn("Triangle includes extraneous development lags")

        obj = self.dev_to_val()

        if ograin_new != ograin_old:
            freq = {"Y": "Y", "S": "2Q"}.get(ograin_new, ograin_new)

            if trailing or (obj.origin.freqstr[-3:] != "DEC" and ograin_old != "M"):
                origin_period_end = self.origin[-1].strftime("%b").upper()
            else:
                origin_period_end = "DEC"

            indices = (
                pd.Series(range(len(self.origin)), index=self.origin)
                .resample("-".join([freq, origin_period_end]))
                .indices
            )

            groups = pd.concat(
                [pd.Series([k] * len(v), index=v) for k, v in indices.items()], axis=0
            ).values

            obj = obj.groupby(groups, axis=2).sum()

            obj.origin_close = origin_period_end
            d_start = pd.Period(
                obj.valuation[0],
                freq=(
                    dgrain_old
                    if dgrain_old == "M"
                    else dgrain_old + obj.origin.freqstr[-4:]
                ),
            ).to_timestamp(how="s")

            if len(obj.ddims) > 1 and obj.origin.to_timestamp(how="s")[0] != d_start:
                addl_ts = (
                    pd.period_range(obj.odims[0], obj.valuation[0], freq=dgrain_old)[
                        :-1
                    ]
                    .to_timestamp()
                    .values
                )
                addl = obj.iloc[..., -len(addl_ts) :] * 0
                addl.ddims = addl_ts
                obj = concat((addl, obj), axis=-1)
                obj.values = num_to_nan(obj.values)

        if dgrain_old != dgrain_new and obj.shape[-1] > 1:
            step = self._dstep()[dgrain_old][dgrain_new]
            d = np.sort(
                len(obj.development) - np.arange(0, len(obj.development), step) - 1
            )

            if obj.is_cumulative:
                obj = obj.iloc[..., d]
            else:
                ddims = obj.ddims[d]
                d2 = [d[0]] * (d[0] + 1) + list(np.repeat(np.array(d[1:]), step))
                obj = obj.groupby(d2, axis=3).sum()
                obj.ddims = ddims

            obj.development_grain = dgrain_new

        obj = obj.dev_to_val() if self.is_val_tri else obj.val_to_dev()

        if inplace:
            self = obj
            return self

        return obj

    def trend(
        self,
        trend=0.0,
        axis="origin",
        start=None,
        end=None,
        ultimate_lag=None,
        **kwargs
    ):
        """Allows for the trending of a Triangle object along either a valuation
        or origin axis.  This method trends using days and assumes a years is
        365.25 days long.

        Parameters
        ----------
        trend : float
            The annual amount of the trend. Use 1/(1+trend)-1 to detrend.
        axis : str (options: ['origin', 'valuation'])
            The axis on which to apply the trend
        start: date
            The start date from which trend should be calculated. If none is
            provided then the latest date of the triangle is used.
        end: date
            The end date to which the trend should be calculated. If none is
            provided then the earliest period of the triangle is used.
        ultimate_lag : int
            If ultimate valuations are in the triangle, optionally set the overall
            age (in months) of the ultimate to be some lag from the latest non-Ultimate
            development

        Returns
        -------
        Triangle
            updated with multiplicative trend applied.
        """
        if axis not in ["origin", "valuation", 2, -2]:
            raise ValueError(
                "Only origin and valuation axes are supported for trending"
            )
        xp = self.get_array_module()
        start = pd.to_datetime(start) if type(start) is str else start
        start = self.valuation_date if start is None else start
        end = pd.to_datetime(end) if type(end) is str else end
        end = self.origin[0].to_timestamp() if end is None else end
        if axis in ["origin", 2, -2]:
            vector = pd.DatetimeIndex(
                np.tile(
                    self.origin.to_timestamp(how="e").values, self.shape[-1]
                ).flatten()
            )
        else:
            vector = self.valuation
        lower, upper = (end, start) if end > start else (start, end)
        vector = pd.DatetimeIndex(
            np.maximum(
                np.minimum(np.datetime64(lower), vector.values), np.datetime64(upper)
            )
        )
        vector = (
            (start.year - vector.year) * 12 + (start.month - vector.month)
        ).values.reshape(self.shape[-2:], order="f")
        if self.is_ultimate and ultimate_lag is not None and vector.shape[-1] > 1:
            vector[:, -1] = vector[:, -2] + ultimate_lag
        trend = (
            xp.array((1 + trend) ** (vector / 12))[None, None, ...] * self.nan_triangle
        )
        obj = self.copy()
        obj.values = obj.values * trend
        return obj

    def broadcast_axis(self, axis, value):
        warnings.warn(
            """
            Broadcast axis is deprecated in favor of broadcasting
            using Triangle arithmetic."""
        )
        return self

    def copy(self):
        X = object.__new__(self.__class__)
        X.__dict__.update(vars(self))
        X._set_slicers()
        X.values = X.values.copy()
        return X

    def development_correlation(self, p_critical=0.5):
        """
        Mack (1997) test for correlations between subsequent development
        factors. Results should be within confidence interval range
        otherwise too much correlation

        Parameters
        ----------
        p_critical: float (default=0.10)
            Value between 0 and 1 representing the confidence level for the test. A
            value of 0.1 implies 90% confidence.
        Returns
        -------
            DevelopmentCorrelation object with t, t_critical, t_expectation,
            t_variance, and range attributes.
        """
        return DevelopmentCorrelation(self, p_critical)

    def valuation_correlation(self, p_critical=0.1, total=False):
        """
        Mack test for calendar year effect
        A calendar period has impact across developments if the probability of
        the number of small (or large) development factors in that period
        occurring randomly is less than p_critical

        Parameters
        ----------
        p_critical: float (default=0.10)
            Value between 0 and 1 representing the confidence level for the test
        total:
            Whether to calculate valuation correlation in total across all
            years (True) consistent with Mack 1993 or for each year separately
            (False) consistent with Mack 1997.
        Returns
        -------
            ValuationCorrelation object with z, z_critical, z_expectation and
            z_variance attributes.

        """
        return ValuationCorrelation(self, p_critical, total)

    def shift(self, periods=-1, axis=3):
        """Shift elements along an axis by desired number of periods.

        Data that falls beyond the existing shape of the Triangle is eliminated
        and new cells default to zero.

        Parameters
        ----------
        periods : int
            Number of periods to shift. Can be positive or negative.
        axis : {2 or 'origin', 3 or 'development', None}, default 3
            Shift direction.

        Returns
        -------
        Triangle
            updated with shifted elements

        """
        axis = self._get_axis(axis)
        if axis < 2:
            raise AttributeError(
                "Lagging only supported for origin and development axes"
            )
        if periods == 0:
            return self
        if periods > 0:
            if axis == 3:
                out = concat(
                    (
                        self.iloc[..., 1:].rename("development", self.development[:-1]),
                        (self.iloc[..., -1:] * 0),
                    ),
                    axis=axis,
                )
            else:
                out = concat(
                    (
                        self.iloc[..., 1:, :].rename("origin", self.origin[:-1]),
                        (self.iloc[..., -1:, :] * 0),
                    ),
                    axis=axis,
                )
        else:
            if axis == 3:
                out = concat(
                    (
                        (self.iloc[..., :1] * 0),
                        self.iloc[..., :-1].rename("development", self.development[1:]),
                    ),
                    axis=axis,
                )
            else:
                out = concat(
                    (
                        (self.iloc[..., :1, :] * 0),
                        self.iloc[..., :-1, :].rename("origin", self.origin[1:]),
                    ),
                    axis=axis,
                )
        if abs(periods) == 1:
            return out
        else:
            return out.shift(periods - 1 if periods > 0 else periods + 1, axis)

    def sort_axis(self, axis):
        """Method to sort a Triangle along a given axis

        Parameters
        ----------
        axis : in or str
            The axis for sorting

        Returns
        -------
        Triangle
            updated with shifted elements
        """

        axis = self._get_axis(axis)
        if axis == 0:
            return self.sort_index()
        obj = self.copy()
        if axis == 1:
            sort = pd.Series(self.vdims).sort_values().index
            if np.any(sort != pd.Series(self.vdims).index):
                obj.values = obj.values[:, list(sort), ...]
                obj.vdims = obj.vdims[list(sort)]
        if axis == 2:
            sort = pd.Series(self.odims).sort_values().index
            if np.any(sort != pd.Series(self.odims).index):
                obj.values = obj.values[..., list(sort), :]
                obj.odims = obj.odims[list(sort)]
        if axis == 3:
            sort = self.development.sort_values().index
            if np.any(sort != self.development.index):
                obj.values = obj.values[..., list(sort)]
                obj.ddims = obj.ddims[list(sort)]
        return obj

    def reindex(self, columns=None, fill_value=np.nan):
        obj = self.copy()
        for column in columns:
            if column not in obj.columns:
                obj[column] = fill_value
        return obj



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
from sklearn.base import BaseEstimator
import json
import joblib
import dill


class TriangleIO:
    def to_pickle(self, path, protocol=None):
        """ Serializes triangle object to pickle.

        Parameters
        ----------
        path : str
            File path and name of pickle object.
        protocol :
            The pickle protocol to use.

        """
        with open(path, "wb") as pkl:
            dill.dump(self, pkl)

    def to_json(self):
        """ Serializes triangle object to json format

        Returns
        -------
            string representation of object in json format
        """
        metadata = {
            "is_val_tri": self.is_val_tri,
            "is_cumulative": self.is_cumulative,
            "is_pattern": self.is_pattern,
            "columns": list(self.columns),
        }
        out = self.cum_to_incr().dev_to_val().to_frame(
            keepdims=True, origin_as_datetime=True).fillna(0)
        x = out.reset_index().to_json(orient="split", date_unit="ns")
        json_dict = {"metadata": json.dumps(metadata), "data": x}
        sub_tris = [k for k, v in vars(self).items() if isinstance(v, TriangleIO)]
        json_dict["sub_tris"] = {
            sub_tri: getattr(self, sub_tri).to_json() for sub_tri in sub_tris
        }
        dfs = [k for k, v in vars(self).items() if isinstance(v, pd.DataFrame)]
        json_dict["dfs"] = {df: getattr(self, df).to_json() for df in dfs}
        dfs = [k for k, v in vars(self).items() if isinstance(v, pd.Series)]
        json_dict["dfs"].update(
            {df: getattr(self, df).to_frame().to_json() for df in dfs}
        )
        return json.dumps(json_dict)


class EstimatorIO:
    """ Class intended to allow persistence of estimator objects """

    def to_pickle(self, path, protocol=None):
        """ Serializes triangle object to pickle.

        Parameters
        ----------
        path : str
            File path and name of pickle object.
        protocol :
            The pickle protocol to use.
        """
        with open(path, "wb") as pkl:
            dill.dump(self, pkl)

    def to_json(self):
        """ Serializes triangle object to json format

        Returns
        -------
            string representation of object in json format
        """
        params = self.get_params(deep=False)
        j = lambda v: v.to_json() if isinstance(v, BaseEstimator) else v
        params = {k: j(v) for k, v in params.items()}
        return json.dumps({"params": params, "__class__": self.__class__.__name__})



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.tails import TailBase
from chainladder.utils import WeightedRegression
from chainladder.development import Development
import pandas as pd
import warnings


class TailCurve(TailBase):
    """Allows for extraploation of LDFs to form a tail factor.

    Parameters
    ----------
    curve : str ('exponential', 'inverse_power')
        The type of curve extrapolation you'd like to use
    fit_period : tuple (start, stop) or list(bool)
        A tuple representing the range of ldfs to use in the curve fit.
        The use of ``None`` will use the edge of the triangle.  For example,
        (48, None) will use development factors for age 48 and beyond.
        Alternatively, passing a list of booleans [True, False, ...] will
        allow for excluding (False) any development patterns from fitting.
    extrap_periods : int
        Then number of development periods from attachment point to extrapolate
        the fit.
    errors : str ('raise' or 'ignore')
        Whether to raise an error or ignore observations that violate the
        distribution being fit.  The most common is ldfs < 1.0 will not work
        in either the ``exponential`` or ``inverse_power`` fits.
    attachment_age: int (default=None)
        The age at which to attach the fitted curve.  If None, then the latest
        age is used. Measures of variability from original ``ldf_`` are retained
        when being used in conjunction with the MackChainladder method.
    reg_threshold : tuple (lower, upper)
        A tuple representing the lower and upper thresholds for the ldfs to be
        considered in the log regression of the tail fitting. Default lower
        threshold set to 1.00001 to avoid distortion caused by ldfs close to 1.
        Upper threshold can be used as an alternative to the fit_period start,
        to make the selection value based rather then period based.
    projection_period : int
        The number of months beyond the latest available development age the
        `ldf_` and `cdf_` vectors should extend.


    Attributes
    ----------
    ldf_ : Triangle
        ldf with tail applied.
    cdf_ : Triangle
        cdf with tail applied.
    tail_ : DataFrame
        Point estimate of tail at latest maturity available in the Triangle.
    slope_ : DataFrame
        Slope parameter of the curve fit.
    intercept : DataFrame
        Intercept parameter of the curve fit.
    """

    def __init__(
            self,
            curve="exponential",
            fit_period=(None, None),
            extrap_periods=100,
            errors="ignore",
            attachment_age=None,
            reg_threshold=(1.00001, None),
            projection_period=12
    ):
        # validate arguments

        if curve not in [
            'exponential',
            'inverse_power'
        ]:
            raise ValueError("Invalid curve type specified. Accepted values are 'exponential' and 'inverse_power'.")

        if errors not in [
            'ignore',
            'raise'
        ]:
            raise ValueError("Invalid value argument supplied to the errors parameter. Accepted values are 'raise' "
                             "and 'ignore'.")
        self.curve = curve
        self.fit_period = fit_period
        self.extrap_periods = extrap_periods
        self.errors = errors
        self.attachment_age = attachment_age
        self.reg_threshold = reg_threshold
        self.projection_period = projection_period

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the tail will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        from chainladder.utils.utility_functions import num_to_nan

        X = X.copy()
        xp = X.get_array_module()
        if type(self.fit_period) == slice:
            warnings.warn(
                "Slicing for fit_period is deprecated and will be removed. Please use a tuple (start_age, end_age)."
            )
            fit_period = self.fit_period
        elif type(self.fit_period) is list:
            fit_period = xp.array(self.fit_period)[None, None, None, :]
        else:
            grain = {"Y": 12, "S": 6, "Q": 3, "M": 1}[X.development_grain]
            start = (
                None
                if self.fit_period[0] is None
                else int(self.fit_period[0] / grain - 1)
            )
            end = (
                None
                if self.fit_period[1] is None
                else int(self.fit_period[1] / grain - 1)
            )
            fit_period = slice(start, end, None)
        super().fit(X, y, sample_weight)
        xp = self.ldf_.get_array_module()
        _y = self.ldf_.values[..., : X.shape[-1] - 1].copy()
        _w = xp.zeros(_y.shape)
        if type(fit_period) is slice:
            _w[..., fit_period] = 1.0
        else:
            _w = (_w + 1) * fit_period
        if self.reg_threshold[0] is None:
            warnings.warn(
                "Lower threshold for ldfs not set. Lower threshold will be set to 1.0 to ensure"
                "valid inputs for regression.")
            lower_threshold = 1
        elif self.reg_threshold[0] < 1:
            warnings.warn(
                "Lower threshold for ldfs set too low (<1). Lower threshold will be set to 1.0 to ensure "
                "valid inputs for regression.")
            lower_threshold = 1
        else:
            lower_threshold = self.reg_threshold[0]
        if self.reg_threshold[1] is not None:
            if self.reg_threshold[1] <= lower_threshold:
                warnings.warn(
                    "Can't set upper threshold for ldfs below lower threshold. Upper threshold will be set to 'None'.")
                upper_threshold = None
            else:
                upper_threshold = self.reg_threshold[1]
        else:
            upper_threshold = self.reg_threshold[1]
        if self.errors == "ignore":
            if upper_threshold is None:
                _w[_y <= lower_threshold] = 0
                _y[_y <= lower_threshold] = 1.01
            else:
                _w[(_y <= lower_threshold) | (_y > upper_threshold)] = 0
                _y[(_y <= lower_threshold) | (_y > upper_threshold)] = 1.01
        elif self.errors == "raise" and xp.any(_y < 1.0):
            raise ZeroDivisionError("Tail fit requires all LDFs to be greater than 1.0")
        _y = xp.log(_y - 1)
        n_obs = X.shape[-1] - 1
        k, v = X.shape[:2]
        _x = self._get_x(_w, _y)
        # Get LDFs
        coefs = WeightedRegression(axis=3, xp=xp).fit(_x, _y, _w)
        self._slope_, self._intercept_ = coefs.slope_, coefs.intercept_
        extrapolate = xp.cumsum(
            xp.ones(tuple(list(_y.shape)[:-1] + [self.extrap_periods + n_obs])), -1
        )
        tail = self._predict_tail(extrapolate)
        if self.attachment_age:
            attach_idx = xp.min(xp.where(X.ddims >= self.attachment_age))
        else:
            attach_idx = len(X.ddims) - 1
        self.ldf_.values = xp.concatenate(
            (self.ldf_.values[..., :attach_idx], tail[..., attach_idx:]), -1
        )
        obj = Development().fit_transform(X) if "ldf_" not in X else X
        self._get_tail_stats(obj)
        return self

    def _get_x(self, w, y):
        # For Exponential decay, no transformation on x is needed
        if self.curve == "exponential":
            return None
        if self.curve == "inverse_power":
            xp = self.ldf_.get_array_module()
            reg = WeightedRegression(3, False, xp=xp).fit(None, y, w).infer_x_w()
            return xp.log(reg.x)

    def _predict_tail(self, extrapolate):
        xp = self.ldf_.get_array_module()
        if self.curve == "exponential":
            tail_ldf = xp.exp(self._slope_ * extrapolate + self._intercept_)
        if self.curve == "inverse_power":
            tail_ldf = xp.exp(self._intercept_) * (extrapolate ** self._slope_)
        return self._get_tail_prediction(tail_ldf)

    @property
    def slope_(self):
        """ Does not work with munich """
        rows = self.ldf_.index.set_index(self.ldf_.key_labels).index
        return pd.DataFrame(
            self._slope_[..., 0, 0], index=rows, columns=self.ldf_.vdims
        )

    @property
    def intercept_(self):
        """ Does not work with munich """
        rows = self.ldf_.index.set_index(self.ldf_.key_labels).index
        return pd.DataFrame(
            self._intercept_[..., 0, 0], index=rows, columns=self.ldf_.vdims
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
from sparse._slicing import normalize_index
from chainladder.utils.utility_functions import num_to_nan

class _LocBase:
    """ Base class for pandas style loc/iloc indexing """

    def __init__(self, obj):
        self.obj = obj

    def get_idx(self, idx):
        """ Returns a slice of the original Triangle """
        obj = self.obj.copy()
        i_idx = _LocBase._contig_slice(idx[0])
        c_idx = _LocBase._contig_slice(idx[1])
        o_idx = _LocBase._contig_slice(idx[2])
        d_idx = _LocBase._contig_slice(idx[3])
        if type(o_idx) != slice or type(d_idx) != slice:
            raise ValueError("Fancy indexing on origin/development is not supported.")
        if type(i_idx) is slice or type(c_idx) is slice:
            obj.values = obj.values[i_idx, c_idx, o_idx, d_idx]
        else:
            obj.values = obj.values[i_idx, :, o_idx, d_idx][:, c_idx, ...]
        obj.kdims = obj.kdims[i_idx]
        obj.vdims = obj.vdims[c_idx]
        obj.odims, obj.ddims = obj.odims[o_idx], obj.ddims[d_idx]
        obj.iloc, obj.loc = Ilocation(obj), Location(obj)
        obj.valuation_date = np.minimum(obj.valuation.max(), obj.valuation_date)
        return obj

    @staticmethod
    def _contig_slice(arr):
        """ Try to make a contiguous slicer from an array of indices """
        if type(arr) is slice:
            return arr
        if type(arr) in [int, np.int64, np.int32]:
            arr = [arr]
        if len(arr) == 1:
            return slice(arr[0], arr[0] + 1)
        diff = np.diff(arr)
        if len(diff) == 0:
            raise ValueError("Slice returns empty Triangle")
        if max(diff) == min(diff):
            step = max(diff)
        else:
            return arr
        step = None if step == 1 else step
        min_arr = None if min(arr) == 0 else min(arr)
        max_arr = max(arr) + 1
        if step and step < 0:
            min_arr, max_arr = max_arr - 1, min_arr - 1 if min_arr else min_arr
        return slice(min_arr, max_arr, step)

    def __setitem__(self, key, values):
        if self.obj.array_backend == "sparse":
            raise ValueError('Setting values with sparse backend requires .at or .iat')
        if isinstance(values, TriangleSlicer):
            values = values.values
        key = tuple(
            [slice(item, item + 1) if type(item) is int else item for item in key]
        )
        self.obj.values.__setitem__(self._normalize_index(key), values)

    def _normalize_index(self, key):
        key = normalize_index(key, self.obj.shape)
        l = []
        for n, i in enumerate(key):
            if type(i) is slice:
                start = i.start if i.start > 0 else None
                stop = i.stop if i.stop > -1 else None
                stop = None if stop == self.obj.shape[n] else stop
                step = None if start is None and stop is None else i.step
                l.append(slice(start, stop, step))
            else:
                l.append(i)
        key = tuple(l)
        return key

    def _sparse_setitem(self, key, values):
        check = (
            (self.obj.values.coords[0] == key[0]) *
            (self.obj.values.coords[1] == key[1]) *
            (self.obj.values.coords[2] == key[2]) *
            (self.obj.values.coords[3] == key[3]))
        if check.max():
            data_index = np.where(check == True)[0][0]
            self.obj.values.data[data_index] = values
        else:
            self.obj.values.coords = np.concatenate(
                (self.obj.values.coords, np.array(key)[:, None]), 1)
            self.obj.values.data = np.concatenate(
                (self.obj.values.data, np.array([values])), 0)
            self.obj.values = self.obj.get_array_module()(
                self.obj.values.coords, self.obj.values.data, prune=True,
                has_duplicates=False, shape=self.obj.shape,
                fill_value=self.obj.values.fill_value)


class Location(_LocBase):
    """ class to generate .loc[] functionality """

    def __getitem__(self, key):
        obj = self.get_idx(self.key_to_slice(key))
        if len(obj) > 1 and obj.index.iloc[:, 0].nunique() == 1:
            obj.set_index(obj.index.iloc[:, 1:], inplace=True)
        return obj

    def format_key(self, key):
        """ Converts keys to loc slices """
        if (type(key) is tuple and len(key) > 1
            and len(self.obj.key_labels) > 1 and type(key[1]) is str
            and key[1] in self.obj.index[self.obj.key_labels[1]].unique()):
            key = (key,)
        else:
            key = (key,) if type(key) is not tuple else key
        key_mask = tuple([i if i is Ellipsis else 0 for i in key])
        if len(key_mask) < len(self.obj.shape) and Ellipsis not in key_mask:
            key_mask = tuple(list(key_mask) + [Ellipsis])
        key_mask = list(self._normalize_index(key_mask))
        key = [item for item in key if item is not Ellipsis]
        for i in range(len(self.obj.shape)):
            if key_mask[i] == 0:
                key_mask[i] = key[0]
                key.pop(0)
        return key_mask

    def index_key(self, key):
        if type(key) == pd.Series and len(key) != len(self.obj):
                key = key.to_frame().T
        if type(key) == pd.Series:
            idx = np.where(key)[0]
        elif type(key) == pd.DataFrame:
            idx = (self.obj.index.reset_index().set_index(self.obj.key_labels)
                       .loc[key.set_index(list(key.columns)).index]).values.flatten()
        elif type(key) in [slice, list, tuple]:
            idx = (self.obj.index.reset_index()
                       .set_index(self.obj.key_labels).loc[key]).values.flatten()
        else:
            idx = np.where(self.obj.kdims[:, 0]==key)[0]
        return idx

    def other_key(self, key, idx):
        if type(key) is np.ndarray:
            return np.where(key)[0]
        if key == slice(None, None, None):
            return key
        s = getattr(self.obj, idx)
        if type(key) in [slice, list]:
            return pd.Series(range(len(s)), index=s).loc[key].values
        if not hasattr(key, '__iter__') or type(key) is str:
            return np.array([pd.Series(range(len(s)), index=s).loc[key]])
        else:
            raise AttributeError("Unable to slice.")


    def key_to_slice(self, key):
        """ Converts keys to integer slices """
        key = self.format_key(key)
        out = [self.index_key(key[0]),
                self.other_key(key[1], 'columns'),
                self.other_key(key[2], 'origin'),
                self.other_key(key[3], 'development')]
        return out

    def __setitem__(self, key, values):
        super().__setitem__(self.key_to_slice(key), values)

class Ilocation(_LocBase):
    """ class to generate .iloc[] functionality """

    def __getitem__(self, key):
        return self.get_idx(self._normalize_index(key))

    def __setitem__(self, key, values):
        super().__setitem__(self._normalize_index(key), values)


class TriangleSlicer:
    def __getitem__(self, key):
        """ Boolean Slicer functionality """
        if type(key) is pd.Series and key.name == "development":
            return self._slice(key, "ddims")
        if type(key) is np.ndarray:
            if len(key) == np.prod(self.shape[-2:]) and self.shape[-1] > 1:
                return self._slice_valuation(key)
            return self._slice(key, "odims")
        if type(key) is pd.Series:
            return self.iloc[self.index[key].index]
        elif key in self.key_labels:
            return self.index[key]
        else:
            if type(key) is str and self.virtual_columns.columns.get(key, None):
                out = self.virtual_columns[key].copy()
                out.virtual_columns.columns = {}
                return out
            key = [key] if not hasattr(key, '__iter__') or type(key) is str else key
            idx = [list(self.vdims).index(item) for item in key]
            return self.iloc[:, idx]

    def __setitem__(self, key, value):
        """ Function for pandas style column setting """
        xp = self.get_array_module()
        if callable(value):
            self.virtual_columns[key] = value
            if self.array_backend == "sparse":
                if key not in self.vdims:
                    k, v, o, d = self.values.shape
                    self.values.shape = k, v + 1, o, d
                    self.vdims = np.append(self.vdims, key)
                return
            value = (self.iloc[:, 0].copy() * xp.nan).set_backend(self.array_backend)
        else:
            self.virtual_columns.pop(key)
        if isinstance(value, TriangleSlicer) and value.array_backend != self.array_backend:
            value = value.set_backend(self.array_backend)
        if key in self.vdims:
            i = np.where(self.vdims == key)[0][0]
            if self.array_backend == "sparse":
                before = self.drop(key).values
                bc = before.coords[1, :]
                before.coords[1] = np.where(bc >= i, bc + 1, bc,)
                value.values.coords[1] = i
                coords = np.concatenate((before.coords, value.values.coords), axis=1)
                data = np.concatenate((before.data, value.values.data))
                self.values = xp(
                    coords, data, shape=self.shape, prune=True, fill_value=xp.nan
                )
            else:
                if isinstance(value, TriangleSlicer):
                    value = value.values
                self.values[:, i : i + 1] = value
        else:
            self.vdims = self.vdims if key in self.vdims else np.append(self.vdims, key)
            try:
                self.values = xp.concatenate((self.values, value.values), axis=1)
            except:
                # For misaligned triangle support
                conc = (self.values, (self.iloc[:, 0] * 0 + value).values)
                self.values = xp.concatenate(conc, axis=1)

    def _slice_valuation(self, key):
        """ private method for handling of valuation slicing """
        obj = self.copy()
        obj.valuation_date = min(obj.valuation[key].max(), obj.valuation_date)
        key = key.reshape(self.shape[-2:], order="f")
        obj.values = num_to_nan(obj.values * obj.get_array_module().array(key))
        return _LocBase(obj).get_idx(
            (slice(None), slice(None),
             np.arange(obj.shape[2])[np.sum(~key, 1) != obj.shape[3]],
             np.arange(obj.shape[3])[np.sum(~key, 0) != obj.shape[2]]))

    def _slice(self, key, axis):
        """ private method for handling of origin/development slicing """
        obj = self.copy()
        setattr(obj, axis, getattr(obj, axis)[key])
        slicer = ..., _LocBase._contig_slice(np.arange(len(key))[key])
        if axis == "odims":
            slicer = tuple(list(slicer) + [slice(None)])
        obj.values = obj.values[slicer]
        return obj

    def _set_slicers(self):
        """ Call any time the shape of index or column changes """
        self.iloc, self.loc = Ilocation(self), Location(self)
        self.iat, self.at = Iat(self), At(self)
        self.virtual_columns = VirtualColumns(self, self.virtual_columns.columns)


class At(Location):
    def _check_index(self, key):
        idx = self.key_to_slice(key)
        for item in idx:
            if type(item) is slice:
                if item.stop - item.start == 1:
                    next
                else:
                    raise ValueError('Invalid Index in At slicer')
            else:
                if len(item) == 1:
                    next
                else:
                    raise ValueError('Invalid Index in At slicer')
        return idx

    def __getitem__(self, key):
        obj = self.get_idx(self._check_index(key))
        if len(obj) > 1 and obj.index.iloc[:, 0].nunique() == 1:
            idx = obj.index.iloc[:, 1:]
            obj.index = idx
        return obj.values[0, 0, 0, 0]

    def __setitem__(self, key, values):
        key = self._check_index(key)
        if self.obj.array_backend == 'sparse':
            key = (key[0][0], key[1][0], key[2][0], key[3][0])
            self._sparse_setitem(key, values)
        else:
            if isinstance(values, TriangleSlicer):
                values = values.values
            key = tuple(
                [slice(item, item + 1) if type(item) is int else item for item in key]
            )
            self.obj.values.__setitem__(self._normalize_index(key), values)


class Iat(Ilocation):
    def _check_index(self, key):
        idx = self._normalize_index(key)
        types = {type(i) for i in idx}
        if len(types) > 1 or list(types)[0] != int:
            raise ValueError('iAt based indexing can only have integer indexers')
        return idx

    def __getitem__(self, key):
        return self.get_idx(self._check_index(key)).values[0, 0, 0, 0]

    def __setitem__(self, key, values):
        key = self._normalize_index(key)
        if self.obj.array_backend == 'sparse':
            self._sparse_setitem(key, values)
        else:
            super().__setitem__(key, values)


class VirtualColumns:
    def __init__(self, triangle, columns=None):
        self.triangle = triangle
        self.columns = {} if not columns else columns

    def __getitem__(self, value):
        return self.columns[value](self.triangle).rename("columns", [value])

    def __setitem__(self, name, value):
        self.columns[name] = value

    def __repr__(self):
        return str(pd.Index(self.columns.keys()))

    def pop(self, key):
        if key in list(self.columns.keys()):
            self.columns.pop(key)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.tails import TailBase
from chainladder.development.clark import ClarkLDF


class TailClark(TailBase):
    """Allows for extraploation of LDFs to form a tail factor.

    .. versionadded:: 0.6.4

    Parameters
    ----------
    growth: {'loglogistic', 'weibull'}
        The growth function to be used in curve fitting development patterns.
        Options are 'loglogistic' and 'weibull'
    truncation_age: int
        The age at which you wish to stop extrapolating development
    attachment_age: int (default=None)
        The age at which to attach the fitted curve.  If None, then the latest
        age is used. Measures of variability from original ``ldf_`` are retained
        when being used in conjunction with the MackChainladder method.
    projection_period: int
        The number of months beyond the latest available development age the
        `ldf_` and `cdf_` vectors should extend.

    Attributes
    ----------
    ldf_:
        ldf with tail applied.
    cdf_:
        cdf with tail applied.
    tail_: DataFrame
        Point estimate of tail at latest maturity available in the Triangle.
    theta_: DataFrame
        Estimates of the theta parameter of the growth curve.
    omega_: DataFrame
        Estimates of the omega parameter of the growth curve.
    elr_: DataFrame
        The Expected Loss Ratio parameter. This only exists when a ``sample_weight``
        is provided to the Estimator.
    scale_: DataFrame
        The scale parameter of the model.
    norm_resid_: Triangle
        The "Normalized" Residuals of the model according to Clark.

    """

    def __init__(self, growth="loglogistic", truncation_age=None,
                 attachment_age=None, projection_period=12):
        self.growth = growth
        self.truncation_age = truncation_age
        self.attachment_age = attachment_age
        self.projection_period = projection_period

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Set of LDFs to which the tail will be applied.
        y: Ignored
        sample_weight : Triangle-like
            Exposure vector used to invoke the Cape Cod method.

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        backend = X.array_backend
        if backend != "numpy":
            X = X.set_backend("numpy")
        else:
            X = X.copy()
        xp = X.get_array_module()
        super().fit(X, y, sample_weight)
        model = ClarkLDF(growth=self.growth).fit(X, sample_weight=sample_weight)
        xp = X.get_array_module()
        age_offset = {"Y": 6.0, "S": 3, "Q": 1.5, "M": 0.5}[X.development_grain]
        fitted = 1 / model.G_(self.ldf_.ddims - age_offset)
        fitted = xp.concatenate(
            (
                fitted.values[..., :-1] / fitted.values[..., 1:],
                fitted.values[..., -1:],
            ),
            -1,
        )
        fitted = xp.repeat(fitted, self.ldf_.values.shape[2], 2)
        attachment_age = self.attachment_age if self.attachment_age else X.ddims[-2]
        self.ldf_.values = xp.concatenate((
            self.ldf_.values[..., : sum(self.ldf_.ddims < attachment_age)],
            fitted[..., -sum(self.ldf_.ddims >= attachment_age) :],),
            axis=-1,)
        self.omega_ = model.omega_
        self.theta_ = model.theta_
        self.G_ = model.G_
        self.scale_ = model.scale_
        self._G = model._G
        self.incremental_fits_ = model.incremental_fits_
        if hasattr(model, "elr_"):
            self.elr_ = model.elr_
        self.norm_resid_ = model.norm_resid_
        if self.truncation_age:
            self.ldf_.values[..., -1:] = self.ldf_.values[..., -1:] * self.G_(self.truncation_age).values
        # self._get_tail_stats(self)
        if backend == "cupy":
            self = self.set_backend("cupy", inplace=True)
        return self

    def transform(self, X):
        """Transform X.

        Parameters
        ----------
        X: Triangle
            Triangle must contain the ``ldf_`` development attribute.

        Returns
        -------
        X_new: Triangle
            New Triangle with tail factor applied to its development
            attributes.
        """
        X_new = super().transform(X)
        triangles = [
            "omega_",
            "theta_",
            "incremental_fits_",
            "G_",
            "_G",
            "growth",
            "scale_",
        ]
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        if hasattr(self, "elr_"):
            X_new.elr_ = self.elr_
        return X_new



from chainladder.adjustments.bootstrap import BootstrapODPSample  # noqa (API import)
from chainladder.adjustments.berqsherm import BerquistSherman  # noqa (API import)
from chainladder.adjustments.parallelogram import ParallelogramOLF  # noqa (API import)
from chainladder.adjustments.trend import Trend  # noqa (API import)
from chainladder.adjustments.trend import TrendConstant  # noqa (API import)



import chainladder as cl
from chainladder.utils.cupy import cp
import numpy as np
import copy
import pandas as pd


def test_non_vertical_line():
    true_olf = (
        1.20
        / (
            (1 - 0.5 * ((31 + 31 + 30 + 31 + 30 + 31) / 365) ** 2) * 1.0
            + (0.5 * ((31 + 31 + 30 + 31 + 30 + 31) / 365) ** 2) * 1.2
        )
        - 1
    )

    result = (
        cl.parallelogram_olf([0.20], ["7/1/2017"], approximation_grain="D")
        .loc["2017"]
        .iloc[0]
        - 1
    )

    assert true_olf == result

    # Monthly approximation
    rate_history = pd.DataFrame(
        {
            "EffDate": ["2010-07-01", "2011-01-01", "2012-07-01", "2013-04-01"],
            "RateChange": [0.035, 0.05, 0.10, -0.01],
        }
    )

    data = pd.DataFrame(
        {"Year": list(range(2006, 2016)), "EarnedPremium": [10_000] * 10}
    )

    prem_tri = cl.Triangle(
        data, origin="Year", columns="EarnedPremium", cumulative=True
    )
    prem_tri = cl.ParallelogramOLF(
        rate_history,
        change_col="RateChange",
        date_col="EffDate",
        approximation_grain="M",
        vertical_line=False,
    ).fit_transform(prem_tri)
    assert (
        np.round(prem_tri.olf_.to_frame().values, 6).flatten()
        == [
            1.183471,
            1.183471,
            1.183471,
            1.183471,
            1.178316,
            1.120181,
            1.075556,
            1.004236,
            0.999684,
            1.000000,
        ]
    ).all()

    # Daily approximation
    rate_history = pd.DataFrame(
        {
            "EffDate": ["2010-07-01", "2011-01-01", "2012-07-01", "2013-04-01"],
            "RateChange": [0.035, 0.05, 0.10, -0.01],
        }
    )

    data = pd.DataFrame(
        {"Year": list(range(2006, 2016)), "EarnedPremium": [10_000] * 10}
    )

    prem_tri = cl.Triangle(
        data, origin="Year", columns="EarnedPremium", cumulative=True
    )
    prem_tri = cl.ParallelogramOLF(
        rate_history,
        change_col="RateChange",
        date_col="EffDate",
        approximation_grain="D",
        vertical_line=False,
    ).fit_transform(prem_tri)
    assert (
        np.round(prem_tri.olf_.to_frame().values, 6).flatten()
        == [
            1.183471,
            1.183471,
            1.183471,
            1.183471,
            1.178231,
            1.120105,
            1.075410,
            1.004073,
            0.999693,
            1.000000,
        ]
    ).all()


def test_vertical_line():
    olf = cl.parallelogram_olf(
        [0.20], ["7/1/2017"], approximation_grain="D", vertical_line=True
    )
    true_olf = 1.2 / ((1 - 184 / 365) * 1.0 + (184 / 365) * 1.2)
    assert abs(olf.loc["2017"].iloc[0] - true_olf) < 0.00001


def test_triangle_json_io(clrd):
    xp = clrd.get_array_module()
    clrd2 = cl.read_json(clrd.to_json(), array_backend=clrd.array_backend)
    xp.testing.assert_array_equal(clrd.values, clrd2.values)
    xp.testing.assert_array_equal(clrd.kdims, clrd2.kdims)
    xp.testing.assert_array_equal(clrd.vdims, clrd2.vdims)
    xp.testing.assert_array_equal(clrd.odims, clrd2.odims)
    xp.testing.assert_array_equal(clrd.ddims, clrd2.ddims)
    assert np.all(clrd.valuation == clrd2.valuation)


def test_json_for_val(raa):
    x = raa.dev_to_val().to_json()
    assert cl.read_json(x) == raa.dev_to_val()


def test_estimator_json_io():
    assert (
        cl.read_json(cl.Development().to_json()).get_params()
        == cl.Development().get_params()
    )


def test_pipeline_json_io():
    pipe = cl.Pipeline(
        steps=[("dev", cl.Development()), ("model", cl.BornhuetterFerguson())]
    )
    pipe2 = cl.read_json(pipe.to_json())
    assert {item[0]: item[1].get_params() for item in pipe.get_params()["steps"]} == {
        item[0]: item[1].get_params() for item in pipe2.get_params()["steps"]
    }


def test_json_subtri(raa):
    a = cl.read_json(cl.Chainladder().fit_predict(raa).to_json()).full_triangle_
    b = cl.Chainladder().fit_predict(raa).full_triangle_
    assert abs(a - b).max().max() < 1e-4


def test_json_df():
    x = cl.MunichAdjustment(paid_to_incurred=("paid", "incurred")).fit_transform(
        cl.load_sample("mcl")
    )
    assert abs(cl.read_json(x.to_json()).lambda_ - x.lambda_).sum() < 1e-5


def test_concat(clrd):
    tri = clrd.groupby("LOB").sum()
    assert (
        cl.concat([tri.loc["wkcomp"], tri.loc["comauto"]], axis=0)
        == tri.loc[["wkcomp", "comauto"]]
    )


def test_model_diagnostics(qtr):
    cl.model_diagnostics(cl.Chainladder().fit(qtr))


def test_concat_immutability(raa):
    u = cl.Chainladder().fit(raa).ultimate_
    l = raa.latest_diagonal
    u.columns = l.columns
    u_new = copy.deepcopy(u)
    cl.concat((l, u), axis=3)
    assert u == u_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
import warnings
from sklearn.utils import deprecated
from chainladder.utils.utility_functions import num_to_nan

try:
    import dask.bag as db
except:
    db = None


class TriangleGroupBy:
    def __init__(self, obj, by, axis=0, **kwargs):
        self.obj = obj.copy()
        self.axis = self.obj._get_axis(axis)
        self.by = by
        if kwargs.get("groups", None):
            self.groups = kwargs.get("groups", None)
        elif self.axis == 0:
            self.groups = obj.index.groupby(by)
        elif axis == 1:
            self.groups = pd.DataFrame(obj.columns).groupby(by)
        else:
            self.groups = pd.DataFrame(by).groupby(by)

    def __getitem__(self, key):
        return TriangleGroupBy(
            obj=self.obj.__getitem__(key),
            by=self.by,
            axis=self.axis,
            groups=self.groups,
        )


class TrianglePandas:
    def to_frame(self, origin_as_datetime=True, keepdims=False,
                 implicit_axis=False, *args, **kwargs):
        """ Converts a triangle to a pandas.DataFrame.
        Parameters
        ----------
        origin_as_datetime : bool
            Whether the origin vector should be converted from PeriodIndex
            into a datetime dtype. Default is False.
        keepdims : bool
            If True, the triangle will be converted to a DataFrame with all
            dimensions intact.  The argument will force a consistent DataFrame
            format regardless of whether any dimensions are of length 1.
        implicit_axis : bool
            When keepdims is True, this denotes whether to include the implicit
            valuation axis in addition to the origin and development.
        Returns
        -------
            pandas.DataFrame representation of the Triangle.
        """
        axes = [num for num, item in enumerate(self.shape) if item > 1]

        if keepdims:
            is_val_tri = self.is_val_tri
            obj = self.val_to_dev().set_backend("sparse")
            out = pd.DataFrame(obj.index.iloc[obj.values.coords[0]])
            out["columns"] = obj.columns[obj.values.coords[1]]
            missing_cols = list(set(self.columns) - set(out['columns']))
            if origin_as_datetime:
                out["origin"] = obj.odims[obj.values.coords[2]]
            else:
                out["origin"] = obj.origin[obj.values.coords[2]]
            out["development"] = obj.ddims[obj.values.coords[3]]
            out["values"] = obj.values.data
            out = pd.pivot_table(
                out, index=obj.key_labels + ["origin", "development"], columns="columns"
            )
            out = out.reset_index().set_index(obj.key_labels)
            out.columns = ["origin", "development"] + list(
                out.columns.get_level_values(1)[2:]
            )

            valuation = pd.DataFrame(
                obj.valuation.values.reshape(obj.shape[-2:], order='F'),
                index=obj.odims if origin_as_datetime else obj.origin, 
                columns=obj.ddims
            ).unstack().rename('valuation').reset_index().rename(
                columns={'level_0': 'development', 'level_1': 'origin'})

            val_dict = dict(zip(list(zip(
                valuation['origin'], valuation['development'])),
                valuation['valuation']))
            if len(out) > 0:
                out['valuation'] = out.apply(
                    lambda x: val_dict[(x['origin'], x['development'])], axis=1)
            else:
                out['valuation'] = self.valuation_date
            col_order = list(self.columns)
            if implicit_axis:
                col_order = ['origin', 'development', 'valuation'] + col_order
            else:
                if is_val_tri:
                    col_order = ['origin', 'valuation'] + col_order
                else:
                    col_order = ['origin', 'development'] + col_order
            for col in set(missing_cols) - self.virtual_columns.columns.keys():
                out[col] = np.nan
            for col in set(missing_cols).intersection(self.virtual_columns.columns.keys()):
                out[col] = out.fillna(0).apply(self.virtual_columns.columns[col], 1)
                out.loc[out[col] == 0, col] = np.nan

            return out[col_order]

        # keepdims = False
        else:
            if self.shape[:2] == (1, 1):
                return self._repr_format(origin_as_datetime)

            elif len(axes) in [1, 2]:
                tri = np.squeeze(self.set_backend("numpy").values)
                axes_lookup = {
                    0: self.kdims,
                    1: self.vdims,
                    2: self.origin,
                    3: self.development,
                }

                if axes[0] == 0:
                    idx = self.index.set_index(self.key_labels).index
                else:
                    idx = axes_lookup[axes[0]]

                if len(axes) == 1:
                    return pd.Series(tri, index=idx).fillna(0)

                elif len(axes) == 2:
                    return pd.DataFrame(
                        tri, index=idx, columns=axes_lookup[axes[1]]
                    ).fillna(0)

            else:
                return self.to_frame(
                    origin_as_datetime=origin_as_datetime, keepdims=True,
                    implicit_axis=implicit_axis)

    def plot(self, *args, **kwargs):
        """Passthrough of pandas functionality"""
        return self.to_frame(origin_as_datetime=False).plot(*args, **kwargs)

    def hvplot(self, *args, **kwargs):
        """Passthrough of pandas functionality"""
        df = self.to_frame(origin_as_datetime=True)
        if type(df.index) == pd.PeriodIndex and len(df.columns) > 1:
            df.index = df.index.to_timestamp(how="s")
        return df.hvplot(*args, **kwargs)

    def _get_axis(self, axis):
        ax = {
            **{0: 0, 1: 1, 2: 2, 3: 3},
            **{-1: 3, -2: 2, -3: 1, -4: 0},
            **{"index": 0, "columns": 1, "origin": 2, "development": 3},
        }
        return ax.get(axis, 0)

    def dropna(self):
        """Method that removes orgin/development vectors from edge of a
        triangle that are all missing values. This may come in handy for a
        new line of business that doesn't have origins/developments of an
        existing line in the same triangle.
        """
        obj = self.sum(axis=0).sum(axis=1)
        xp = obj.get_array_module()
        odim = list((xp.nansum(obj.values[0, 0, :], -1) != 0).astype("int"))
        min_odim = obj.origin[odim.index(1)]
        max_odim = obj.origin[::-1][odim[::-1].index(1)]
        if obj.shape[-1] != 1:
            ddim = list(
                (xp.nansum(obj.values[0, 0, :], -2) != 0).astype("int"))
            ddim = obj.development[pd.Series(ddim).astype(bool)]
            obj = self[
                (self.development >= ddim.min()) & (
                    self.development <= ddim.max())
            ]
            return obj[(self.origin >= min_odim) & (self.origin <= max_odim)]
        obj = self[(self.origin >= min_odim) & (self.origin <= max_odim)]
        return obj

    def fillna(self, value=None, inplace=False):
        """Fill nan with 'value' by axis.
        Parameters
        ----------
        value: single value or array-like values, default = None
            Value(s) to fill across the axis.

        inplace: boolean, default = False
            Whether to modify the triangle object directly (True), or
            return a new modified triangle (False).

        Returns
        -------
        Triangle
        """
        if inplace:
            frame = self + value * 0
            xp = self.get_array_module()
            fill = (xp.nan_to_num(frame.values) == 0) * (self * 0 + value)
            self.values = (frame + fill).values
            return self
        else:
            new_obj = self.copy()
            return new_obj.fillna(value=value, inplace=True)

    def drop(self, labels=None, axis=1):
        """Drop specified labels from rows or columns.

        Remove rows or columns by specifying label names and corresponding axis,
        or by specifying directly index or column names.

        Parameters
        -----------

        label: single label or list-like
            Index or column labels to drop.

        axis: {0 or ‘index’, 1 or ‘columns’}, default 1
            Whether to drop labels from the index (0 or ‘index’)
            or columns (1 or ‘columns’).

        Returns
        -------
        Triangle

        """
        labels = [labels] if type(labels) is str else list(labels)
        if axis == 1:
            return self[[item for item in self.columns if item not in labels]]
        else:
            raise NotImplementedError("drop only inpemented for column axis")

    @property
    def T(self):
        return self.to_frame(origin_as_datetime=False).T

    def groupby(self, by, axis=0, *args, **kwargs):
        """Group Triangle by index values.  If the triangle is convertable to a
        DataFrame, then it defaults to pandas groupby functionality.

        Parameters
        ----------
        by: str or list
            The index to group by

        Returns
        -------
            GroupBy object (pandas or Triangle)
        """
        return TriangleGroupBy(self, by, axis)

    def append(self, other):
        """Append rows of other to the end of caller, returning a new object.

        Parameters
        ----------
        other : Triangle
            The data to append.

        Returns
        -------
            New Triangle with appended data.
        """
        from chainladder.utils.utility_functions import concat

        return concat((self, other), 0)

    def rename(self, axis, value):
        """Alter axes labels.

        Parameters
        ----------
        axis: str or int
            A value of 0 <= axis <= 4 corresponding to axes 'index',
            'columns', 'origin', 'development' respectively.  Both the
            int and str representation can be used.
        value: list or str
            List of new labels to be assigned to the axis. List must be of
            same length of the specified axis.

        Returns
        -------
            Triangle with relabeled axis.
        """
        value = [value] if type(value) is str else value
        if axis == "index" or axis == 0:
            self.index = value
        if axis == "columns" or axis == 1:
            self.columns = value
        if axis == "origin" or axis == 2:
            self.origin = value
        if axis == "development" or axis == 3:
            self.development = value
        return self

    def astype(self, dtype, inplace=True):
        """Copy of the array, cast to a specified type.

        Parameters
        ----------
        dtype : str or dtype
            Typecode or data-type to which the array is cast.
        copy : bool, optional
            By default, astype always returns a newly allocated array.

        Returns
        -------
        Triangle as new datatype
        """
        obj = self.copy() if inplace is False else self
        obj.values = obj.values.astype(dtype)
        return obj

    def head(self, n=5):
        return self.iloc[:n]

    def tail(self, n=5):
        return self.iloc[-n:]

    def sort_index(self, *args, **kwargs):
        return self.iloc[self.index.sort_values(self.key_labels, *args, **kwargs).index]

    def exp(self):
        return self.get_array_module().exp(self)

    def log(self):
        return self.get_array_module().log(self)

    def minimum(self, other):
        return self.get_array_module().minimum(self, other)

    def maximum(self, other):
        return self.get_array_module().maximum(self, other)

    def sqrt(self):
        return self.get_array_module().sqrt(self)

    def round(self, decimals=0, *args, **kwargs):
        return round(self, decimals)


def add_triangle_agg_func(cls, k, v):
    """Aggregate Overrides in Triangle"""

    def agg_func(self, axis=None, *args, **kwargs):
        keepdims = kwargs.get("keepdims", None)
        obj = self.copy()
        auto_sparse = kwargs.pop("auto_sparse", True)
        if axis is None:
            if max(obj.shape) == 1:
                axis = 0
            else:
                axis = min([num for num, _ in enumerate(obj.shape) if _ != 1])
        else:
            axis = self._get_axis(axis)
        xp = obj.get_array_module()
        func = getattr(xp, v)
        kwargs.update({"keepdims": True})
        obj.values = func(obj.values, axis=axis, *args, **kwargs)
        if axis == 0 and obj.values.shape[axis] == 1 and len(obj.kdims) > 1:
            obj.kdims = np.array([["(All)"] * len(obj.key_labels)])
        if axis == 1 and obj.values.shape[axis] == 1 and len(obj.vdims) > 1:
            obj.vdims = np.array([0])
        if axis == 2 and obj.values.shape[axis] == 1 and len(obj.odims) > 1:
            obj.odims = obj.odims[0:1]
        if axis == 3 and obj.values.shape[axis] == 1 and len(obj.ddims) > 1:
            obj.ddims = pd.DatetimeIndex(
                [self.valuation_date], dtype="datetime64[ns]", freq=None
            )
        obj._set_slicers()
        if auto_sparse:
            obj = obj._auto_sparse()
        obj.values = num_to_nan(obj.values)
        if not keepdims and obj.shape == (1, 1, 1, 1):
            return obj.values[0, 0, 0, 0]
        else:
            return obj

    set_method(cls, agg_func, k)


def add_groupby_agg_func(cls, k, v):
    """Aggregate Overrides in GroupBy"""

    def agg_func(self, *args, **kwargs):
        from chainladder.utils import concat
        xp = self.obj.get_array_module()
        obj = self.obj.copy()
        auto_sparse = kwargs.pop("auto_sparse", True)
        if db and obj.array_backend == "sparse":

            def aggregate(i, obj, axis, v):
                return getattr(
                    obj.iloc.__getitem__(tuple([slice(None)] * axis + [i])), v
                )(axis, auto_sparse=False, keepdims=True)

            bag = db.from_sequence(self.groups.indices.values())
            bag = bag.map(aggregate, obj, self.axis, v)
            values = bag.compute(scheduler="threads")
        else:
            values = [
                getattr(
                    obj.iloc.__getitem__(
                        tuple([slice(None)] * self.axis + [i])), v
                )(self.axis, auto_sparse=False, keepdims=True)
                for i in self.groups.indices.values()
            ]
        obj = concat(values, axis=self.axis, ignore_index=True)
        group_index = self.groups.first().index
        if self.axis == 0:
            if isinstance(group_index, pd.MultiIndex):
                index = (
                    pd.DataFrame(
                        np.zeros(len(group_index)),
                        index=group_index,
                        columns=["_"],
                    )
                    .reset_index()
                    .iloc[:, :-1]
                )
                obj.index = index
            else:
                index = pd.DataFrame(group_index)
                obj.key_labels = index.columns.tolist()
                obj.kdims = index.values
        if self.axis == 1:
            obj.vdims = pd.DataFrame(group_index).values[:, 0]
        if self.axis == 2:
            odims = self.obj._to_datetime(
                pd.Series(self.groups.indices.keys()).to_frame(), [0]
            )
            obj.origin_grain = self.obj._get_grain(odims)
            split = obj.origin_grain.split("-")
            obj.origin_grain = {"A": "Y", "2Q": "S"}.get(split[0], split[0])
            obj.odims = odims.values
        obj._set_slicers()
        if auto_sparse:
            obj = obj._auto_sparse()
        return obj

    set_method(cls, agg_func, k)


def add_df_passthru(cls, k):
    """Pass Through of pandas functionality"""

    def df_passthru(self, *args, **kwargs):
        return getattr(pd.DataFrame, k)(self.to_frame(), *args, **kwargs)

    set_method(cls, df_passthru, k)


def set_method(cls, func, k):
    """Assigns methods to a class"""
    func.__doc__ = "Refer to pandas for ``{}`` functionality.".format(k)
    func.__name__ = k
    setattr(cls, func.__name__, func)


df_passthru = (
    [
        "to_clipboard",
        "to_csv",
        "to_excel",
        "to_json",
        "to_html",
    ]
    + [
        "to_dict",
        "unstack",
        "pivot",
        "drop_duplicates",
        "describe",
        "melt",
    ]
    + [
        "pct_chg",
    ]
)
for item in df_passthru:
    add_df_passthru(TrianglePandas, item)

agg_funcs = ["sum", "mean", "median", "max", "min", "prod", "var"]
agg_funcs = agg_funcs + ["std", "cumsum", "quantile"]
for k in agg_funcs:
    add_groupby_agg_func(TriangleGroupBy, k, k)
agg_funcs = {item: "nan" + item for item in agg_funcs}
more_aggs = ["diff"]
agg_funcs = {**agg_funcs, **{item: item for item in more_aggs}}
for k, v in agg_funcs.items():
    add_triangle_agg_func(TrianglePandas, k, v)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.tails import TailBase
from chainladder.development import Development


class TailConstant(TailBase):
    """Allows for the entry of a constant tail factor to LDFs.

    Parameters
    ----------
    tail: float
        The constant to apply to all LDFs within a triangle object.
    decay: float (default=0.50)
        An exponential decay constant that allows for decay over future
        development periods.  A decay rate of 0.5 sets the development portion
        of each successive LDF to 50% of the previous LDF.
    attachment_age: int (default=None)
        The age at which to attach the fitted curve.  If None, then the latest
        age is used. Measures of variability from original ``ldf_`` are retained
        when being used in conjunction with the MackChainladder method.
    projection_period: int
        The number of months beyond the latest available development age the
        `ldf_` and `cdf_` vectors should extend.

    Attributes
    ----------
    ldf_:
        ldf with tail applied.
    cdf_:
        cdf with tail applied.
    tail_: DataFrame
        Point estimate of tail at latest maturity available in the Triangle.

    Notes
    -----
    The tail constant does not support the entry of variability parameters
    necessary for stochastic approaches, so any usage of TailConstant will be
    inherently deterministic.

    See also
    --------
    TailCurve

    """

    def __init__(self, tail=1.0, decay=0.5, attachment_age=None, projection_period=12):
        self.tail = tail
        self.decay = decay
        self.attachment_age = attachment_age
        self.projection_period = projection_period

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the tail will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        xp = self.ldf_.get_array_module()
        tail = self.tail
        if self.attachment_age:
            attach_idx = xp.min(xp.where(X.ddims >= self.attachment_age))
        else:
            attach_idx = len(X.ddims) - 1
        self = self._apply_decay(X, tail, attach_idx)
        obj = Development().fit_transform(X) if "ldf_" not in X else X
        self._get_tail_stats(obj)
        return self



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.methods import MethodBase


class Chainladder(MethodBase):
    """
    The basic determinsitic chainladder method.

    Parameters
    ----------
    None

    Attributes
    ----------
    X_:
        returns **X** used to fit the triangle
    ultimate_:
        The ultimate losses per the method
    ibnr_:
        The IBNR per the method
    full_expectation_:
        The ultimates back-filled to each development period in **X** replacing
        the known data
    full_triangle_:
        The ultimates back-filled to each development period in **X** retaining
        the known data
    """

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Data to which the model will be applied.
        y: Ignored
        sample_weight : Ignored

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        self.ultimate_ = self._get_ultimate(self.X_)
        self.process_variance_ = self._include_process_variance()
        return self

    def predict(self, X, sample_weight=None):
        """Predicts the chainladder ultimate on a new triangle **X**

        Parameters
        ----------
        X: Triangle
            Loss data to which the model will be applied.
        sample_weight: Triangle
            Required exposure to be used in the calculation.

        Returns
        -------
        X_new: Triangle
            Loss data with chainladder ultimate applied
        """
        X_new = super().predict(X, sample_weight)
        X_new.ultimate_ = self._get_ultimate(X_new, sample_weight)
        return X_new

    def _get_ultimate(self, X, sample_weight=None):
        """ Private method that uses CDFs to obtain an ultimate vector """
        if X.is_cumulative == False:
            ld = X.sum('development')
            ultimate = ld.val_to_dev()
        else:
            ld = X.latest_diagonal
            ultimate = X.copy()
        cdf = self._align_cdf(ultimate, sample_weight) 
        ultimate = ld * cdf 
        return self._set_ult_attr(ultimate)



import chainladder as cl

def test_mack_to_triangle():
    assert (
        cl.MackChainladder()
        .fit(
            cl.TailConstant().fit_transform(
                cl.Development().fit_transform(cl.load_sample("ABC"))
            )
        )
        .summary_
        == cl.MackChainladder()
        .fit(cl.Development().fit_transform(cl.load_sample("ABC")))
        .summary_
    )


def test_mack_malformed():
    a  = cl.load_sample('raa')
    b = a.iloc[:, :, :-1]
    x = cl.MackChainladder().fit(a) 
    y = cl.MackChainladder().fit(b)
    assert x.process_risk_.iloc[:,:,:-1] == y.process_risk_


import chainladder as cl
import numpy as np


def test_preserve_diagonal():
    triangle = cl.load_sample("berqsherm").loc["Auto"]
    xp = triangle.get_array_module()
    berq = cl.BerquistSherman(
        paid_amount="Paid",
        incurred_amount="Incurred",
        reported_count="Reported",
        closed_count="Closed",
    )
    berq_triangle = berq.fit_transform(triangle)
    assert (
        xp.nansum((berq_triangle.latest_diagonal - triangle.latest_diagonal).values)
        == 0
    )
    assert berq_triangle != triangle

def test_adjusted_values():
    triangle = cl.load_sample("berqsherm").loc["MedMal"]
    xp = triangle.get_array_module()
    berq = cl.BerquistSherman(
        paid_amount="Paid",
        incurred_amount="Incurred",
        reported_count="Reported",
        closed_count="Closed",
        trend=0.15,
    )
    berq_triangle = berq.fit_transform(triangle)

    assert np.allclose(
        triangle["Reported"].values, berq_triangle["Reported"].values, equal_nan=True
    )

    # Ensure that the incurred, paid, and closed count columns are as expected
    berq_triangle.values[np.isnan(berq_triangle.values)] = 0
    assert np.isclose(
        berq_triangle["Incurred"].values.sum(), 1126985253.661, atol=1e-2
    )
    assert np.isclose(berq_triangle["Paid"].values.sum(), 182046766.054, atol=1e-2)
    assert np.isclose(berq_triangle["Closed"].values.sum(), 8798.982, atol=1e-2)


import numpy as np
import chainladder as cl

raa = cl.load_sample("RAA")
raa_1989 = raa[raa.valuation < raa.valuation_date]
raa_1990 = raa[raa.origin > "1981"]
raa_1990 = raa_1990[raa_1990.valuation <= raa_1990.valuation_date]
cl_ult = cl.Chainladder().fit(raa).ultimate_  # Chainladder Ultimate
apriori = cl_ult * 0 + (float(cl_ult.sum()) / 10)  # Mean Chainladder Ultimate
apriori_1989 = apriori[apriori.origin < "1990"]
apriori_1990 = apriori[apriori.origin > "1981"]


def test_voting_predict():
    bcl = cl.Chainladder()
    bf = cl.BornhuetterFerguson()
    cc = cl.CapeCod()

    estimators = [('bcl', bcl), ('bf', bf), ('cc', cc)]
    weights = np.array([[1, 2, 3]] * 3 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3)

    vot = cl.VotingChainladder(
            estimators=estimators,
            weights=weights
        ).fit(
            raa_1989,
            sample_weight=apriori_1989,
        )
    vot.predict(raa_1990, sample_weight=apriori_1990)



import chainladder as cl
import pandas as pd
import numpy as np
import pytest
import io
from datetime import datetime

try:
    from IPython.core.display import HTML
except:
    HTML = None


def test_repr(raa):
    np.testing.assert_array_equal(
        pd.read_html(raa._repr_html_())[0].set_index("Unnamed: 0").values,
        raa.to_frame(origin_as_datetime=False).values,
    )


def test_to_frame_unusual(clrd):
    a = (
        clrd.groupby(["LOB"])
        .sum()
        .latest_diagonal["CumPaidLoss"]
        .to_frame(origin_as_datetime=False)
    )
    b = (
        clrd.latest_diagonal["CumPaidLoss"]
        .groupby(["LOB"])
        .sum()
        .to_frame(origin_as_datetime=False)
    )
    assert (a == b).all().all()


def test_link_ratio(raa, atol):
    assert (
        raa.link_ratio * raa.iloc[:, :, :-1, :-1].values - raa.values[:, :, :-1, 1:]
    ).sum().sum() < atol


def test_incr_to_cum(clrd):
    clrd.cum_to_incr().incr_to_cum() == clrd


def test_create_new_value(clrd):
    clrd2 = clrd.copy()
    clrd2["lr"] = clrd2["CumPaidLoss"] / clrd2["EarnedPremDIR"]
    assert (
        clrd.shape[0],
        clrd.shape[1] + 1,
        clrd.shape[2],
        clrd.shape[3],
    ) == clrd2.shape


def test_multilevel_index_groupby_sum1(clrd):
    assert clrd.groupby("LOB").sum().sum() == clrd.sum()


def test_multilevel_index_groupby_sum2(clrd):
    a = clrd.groupby("GRNAME").sum().sum()
    b = clrd.groupby("LOB").sum().sum()
    assert a == b


def test_boolean_groupby_eq_groupby_loc(clrd):
    assert (
        clrd[clrd["LOB"] == "ppauto"].sum() == clrd.groupby("LOB").sum().loc["ppauto"]
    )


def test_latest_diagonal_two_routes(clrd):
    assert (
        clrd.latest_diagonal.sum()["BulkLoss"] == clrd.sum().latest_diagonal["BulkLoss"]
    )


def test_sum_of_diff_eq_diff_of_sum(clrd):
    assert (clrd["BulkLoss"] - clrd["CumPaidLoss"]).latest_diagonal == (
        clrd.latest_diagonal["BulkLoss"] - clrd.latest_diagonal["CumPaidLoss"]
    )


def test_append(raa):
    raa2 = raa.copy()
    raa2.kdims = np.array([["P2"]])
    raa.append(raa2).sum() == raa * 2
    assert raa.append(raa2).sum() == 2 * raa


def test_assign_existing_col(qtr):
    out = qtr.copy()
    before = out.shape
    out["paid"] = 1 / out["paid"]
    assert out.shape == before


def test_off_cycle_val_date(qtr):
    assert qtr.valuation_date.strftime("%Y-%m-%d") == "2006-03-31"


def test_printer(raa):
    print(raa)


def test_value_order(clrd):
    a = clrd[["CumPaidLoss", "BulkLoss"]]
    b = clrd[["BulkLoss", "CumPaidLoss"]]
    xp = a.get_array_module()
    xp.testing.assert_array_equal(a.values[:, -1], b.values[:, 0])


def test_trend(raa, atol):
    assert abs((raa.trend(0.05).trend((1 / 1.05) - 1) - raa).sum().sum()) < 1e-5


def test_shift(qtr):
    x = qtr.iloc[0, 0]
    xp = x.get_array_module()
    xp.testing.assert_array_equal(x[x.valuation <= x.valuation_date].values, x.values)


def test_quantile_vs_median(clrd):
    xp = clrd.get_array_module()
    xp.testing.assert_array_equal(
        clrd.quantile(q=0.5)["CumPaidLoss"].values, clrd.median()["CumPaidLoss"].values
    )


def test_base_minimum_exposure_triangle(raa):
    d = (
        (raa.latest_diagonal * 0 + 50000)
        .to_frame(origin_as_datetime=False)
        .reset_index()
    )
    d["index"] = d["index"].astype(str)
    cl.Triangle(d, origin="index", columns=d.columns[-1])


def test_origin_and_value_setters(raa):
    raa2 = raa.copy()
    raa.columns = list(raa.columns)
    raa.origin = list(raa.origin)
    assert np.all(
        (
            np.all(raa2.origin == raa.origin),
            np.all(raa2.development == raa.development),
            np.all(raa2.odims == raa.odims),
            np.all(raa2.vdims == raa.vdims),
        )
    )


def test_valdev1(qtr):
    assert qtr.dev_to_val().val_to_dev() == qtr


def test_valdev2(qtr):
    a = qtr.dev_to_val().grain("OYDY").val_to_dev()
    b = qtr.grain("OYDY")
    assert a == b


def test_valdev3(qtr):
    a = qtr.grain("OYDY").dev_to_val().val_to_dev()
    b = qtr.grain("OYDY")
    assert a == b


# def test_valdev4():
#    # Does not work with pandas 0.23, consider requiring only pandas>=0.24
#    raa = raa
#    np.testing.assert_array_equal(raa.dev_to_val()[raa.dev_to_val().development>='1989'].values,
#        raa[raa.valuation>='1989'].dev_to_val().values)


def test_valdev5(raa):
    xp = raa.get_array_module()
    xp.testing.assert_array_equal(
        raa[raa.valuation >= "1989"].latest_diagonal.values, raa.latest_diagonal.values
    )


def test_valdev6(raa):
    xp = raa.get_array_module()
    xp.testing.assert_array_equal(
        raa.grain("OYDY").latest_diagonal.set_backend("numpy").values,
        raa.latest_diagonal.grain("OYDY").set_backend("numpy").values,
    )


def test_valdev7(qtr, atol):
    xp = qtr.get_array_module()
    x = cl.Chainladder().fit(qtr).full_expectation_
    assert xp.sum(x.dev_to_val().val_to_dev().values - x.values) < atol


def test_reassignment(clrd):
    clrd = clrd.copy()
    clrd["values"] = clrd["CumPaidLoss"]
    clrd["values"] = clrd["values"] + clrd["CumPaidLoss"]


def test_dropna(clrd):
    assert clrd.shape == clrd.dropna().shape
    a = clrd[clrd["LOB"] == "wkcomp"].iloc[-5]["CumPaidLoss"].dropna().shape
    assert a == (1, 1, 2, 2)


def test_exposure_tri():
    x = cl.load_sample("auto")
    x = x[x.development == 12]
    x = x["paid"].to_frame(origin_as_datetime=False).T.unstack().reset_index()
    x.columns = ["LOB", "origin", "paid"]
    x.origin = x.origin.astype(str)
    y = cl.Triangle(x, origin="origin", index="LOB", columns="paid")
    x = cl.load_sample("auto")["paid"]
    x = x[x.development == 12]
    assert x == y


def test_jagged_1_add(raa):
    raa1 = raa[raa.origin <= "1984"]
    raa2 = raa[raa.origin > "1984"]
    assert raa2 + raa1 == raa
    assert raa2.dropna() + raa1.dropna() == raa


def test_jagged_2_add(raa):
    raa1 = raa[raa.development <= 48]
    raa2 = raa[raa.development > 48]
    assert raa2 + raa1 == raa
    assert raa2.dropna() + raa1.dropna() == raa


def test_df_period_input(raa):
    d = raa.latest_diagonal
    df = d.to_frame(origin_as_datetime=False).reset_index()
    assert cl.Triangle(df, origin="index", columns=df.columns[-1]) == d


def test_trend_on_vector(raa):
    d = raa.latest_diagonal
    assert (
        d.trend(0.05, axis=2).to_frame(origin_as_datetime=False).astype(int).iloc[0, 0]
        == 29217)


def test_latest_diagonal_val_to_dev(raa):
    assert raa.latest_diagonal.val_to_dev() == raa[raa.valuation == raa.valuation_date]


def test_sumdiff_to_diffsum(clrd):
    out = clrd["CumPaidLoss"]
    assert out.cum_to_incr().incr_to_cum().sum() == out.sum()


def test_init_vector(raa):
    a = raa.latest_diagonal
    b = pd.DataFrame(
        {"AccYear": [item for item in range(1981, 1991)], "premium": [3000000] * 10}
    )
    b = cl.Triangle(b, origin="AccYear", columns="premium")
    assert np.all(a.valuation == b.valuation)
    assert a.valuation_date == b.valuation_date


def test_groupby_axis1(clrd, prism):
    clrd = clrd.sum("origin").sum("development")
    groups = [i.find("Loss") >= 0 for i in clrd.columns]
    assert np.all(
        clrd.to_frame(origin_as_datetime=False).groupby(groups, axis=1).sum()
        == clrd.groupby(groups, axis=1).sum().to_frame(origin_as_datetime=False)
    )
    assert np.all(
        clrd.to_frame(origin_as_datetime=False).groupby("LOB").sum()
        == clrd.groupby("LOB").sum().to_frame(origin_as_datetime=False)
    )
    prism.sum().grain("OYDY")


def test_partial_year(prism):
    before = prism["Paid"].sum().incr_to_cum()
    before = before[before.valuation <= "2017-08"].latest_diagonal

    after = cl.Triangle(
        before.to_frame(keepdims=True, origin_as_datetime=True).reset_index(),
        origin="origin",
        development="valuation",
        columns="Paid",
        index=before.key_labels,
    )

    assert after.valuation_date == before.valuation_date


def test_array_protocol(raa, clrd):
    assert np.sqrt(raa) == raa.sqrt()
    assert np.concatenate((clrd.iloc[:200], clrd.iloc[200:]), 0) == cl.concat(
        (clrd.iloc[:200], clrd.iloc[200:]), 0
    )


# def test_dask_backend(raa):
#     """ Dask backend not fully implemented """
#    raa1 = cl.Chainladder().fit(raa.set_backend('dask')).ultimate_
#    raa2 = cl.Chainladder().fit(raa).ultimate_
#    assert (raa1 == raa2).compute()


def test_partial_val_dev(raa):
    raa = raa.latest_diagonal
    raa.iloc[..., -3:, :] = np.nan
    raa.val_to_dev().iloc[0, 0, 0, -1] == raa.iloc[0, 0, 0, -1]


def test_sort_axis(clrd):
    assert clrd.iloc[::-1, ::-1, ::-1, ::-1].sort_axis(0).sort_axis(1).sort_axis(
        2
    ).sort_axis(3) == clrd.sort_axis(1)


def test_shift(raa):
    assert (
        raa.iloc[..., 1:-1, 1:-1]
        - raa.shift(-1, axis=2)
        .shift(-1, axis=3)
        .shift(2, axis=2)
        .shift(2, axis=3)
        .dropna()
        .values
    ).to_frame(origin_as_datetime=False).fillna(0).sum().sum() == 0


def test_array_protocol2(raa):
    import numpy as np

    assert raa.log().exp() == np.exp(np.log(raa))


def test_create_full_triangle(raa):
    a = cl.Chainladder().fit(raa).full_triangle_
    b = cl.Triangle(
        a.to_frame(keepdims=True, implicit_axis=True, origin_as_datetime=True),
        origin="origin",
        development="valuation",
        columns="values",
    )
    assert a == b


def test_groupby_getitem(clrd):
    assert (
        clrd.groupby("LOB")["CumPaidLoss"].sum()
        == clrd["CumPaidLoss"].groupby("LOB").sum()
    )


def test_virtual_column(prism):
    prism["P"] = prism["Paid"]
    prism["Paid"] = lambda x: x["P"]
    assert prism["Paid"] == prism["P"]


def test_correct_valutaion(raa):
    new = cl.Triangle(
        raa.iloc[..., :-3, :].latest_diagonal.to_frame(
            keepdims=True, implicit_axis=True, origin_as_datetime=True
        ),
        origin="origin",
        development="valuation",
        columns="values",
    )
    assert new.valuation_date == raa.valuation_date


@pytest.mark.xfail
@pytest.mark.parametrize(
    "prop", ["cdf_", "ibnr_", "full_expectation_", "full_triangle_"]
)
def test_no_fitted(raa, prop):
    getattr(raa, prop)


def test_pipe(raa):
    def f(x):
        return x.loc[..., 48:]

    assert raa.loc[..., 48:] == raa.pipe(f)


def test_repr_html(raa, clrd):
    assert type(raa._repr_html_()) == str
    assert type(clrd._repr_html_()) == str


@pytest.mark.xfail(HTML is None, reason="ipython needed for test")
def test_heatmap(raa):
    raa.link_ratio.heatmap()


def test_agg_sparse():
    a = cl.load_sample("raa")
    b = cl.load_sample("raa").set_backend("sparse")
    assert a.mean().mean() == b.mean().mean()


def test_inplace(raa):
    t = raa.copy()
    t.dev_to_val(inplace=True)
    t.val_to_dev(inplace=True)
    t.grain("OYDY", inplace=True)


def test_malformed_init():
    assert (
        cl.Triangle(
            data=pd.DataFrame(
                {
                    "Accident Date": [
                        "2020-07-23",
                        "2019-07-23",
                        "2018-07-23",
                        "2016-07-23",
                        "2020-08-23",
                        "2019-09-23",
                        "2018-10-23",
                    ],
                    "Valuation Date": [
                        "2021-01-01",
                        "2021-01-01",
                        "2021-01-01",
                        "2021-01-01",
                        "2021-01-01",
                        "2021-01-01",
                        "2021-01-01",
                    ],
                    "Loss": [10000, 10000, 10000, 10000, 0, 0, 0],
                }
            ),
            origin="Accident Date",
            development="Valuation Date",
            columns="Loss",
        ).origin_grain
        == "M"
    )


def test_sparse_reassignment_no_mutate(prism):
    x = prism["Paid"].incr_to_cum()
    x["Capped 100k Paid"] = cl.minimum(x["Paid"], 100000)
    x["Excess 100k Paid"] = x["Paid"] - x["Capped 100k Paid"]
    a = x["Excess 100k Paid"].sum().grain("OYDY")
    x["Capped 100k Paid"] = cl.minimum(x["Paid"], 100000)
    x["Excess 100k Paid"] = x["Paid"] - x["Capped 100k Paid"]
    b = x["Excess 100k Paid"].sum().grain("OYDY")
    assert a == b


def test_trailing_origin():
    raa = (
        cl.load_sample("raa")
        .dev_to_val()
        .to_frame(keepdims=True, origin_as_datetime=True)
        .reset_index()
    )
    # adjust valuations to mid-year
    raa["valuation"] = raa["valuation"] - pd.DateOffset(months=6)
    tri = cl.Triangle(
        raa, origin="origin", development="valuation", columns="values", cumulative=True
    )
    assert tri.development.to_list() == [6, 18, 30, 42, 54, 66, 78, 90, 102, 114]
    assert tri.origin_close == "DEC"
    raa["origin2"] = raa["origin"] - pd.DateOffset(months=6)
    tri = cl.Triangle(
        raa,
        origin="origin2",
        development="valuation",
        columns="values",
        cumulative=True,
    )
    assert tri.development.to_list() == [12, 24, 36, 48, 60, 72, 84, 96, 108, 120]
    assert tri.origin_close == "JUN"


def test_trailing_valuation():
    data = (
        cl.load_sample("raa")
        .dev_to_val()
        .to_frame(keepdims=True, origin_as_datetime=True)
    )
    data.valuation = (data.valuation.dt.year + 1) * 100 + 3
    tri = cl.Triangle(data, origin="origin", development="valuation", columns="values")
    assert tri.development.to_list() == [3, 15, 27, 39, 51, 63, 75, 87, 99, 111, 123]
    tri2 = cl.Triangle(
        data, origin="origin", development="valuation", columns="values", trailing=True
    )
    assert tri == tri2


def test_edgecase_236():
    assert (
        cl.Triangle(
            pd.DataFrame(
                {
                    "origin": [201906, 201907],
                    "development": [201911, 201911],
                    "amount": [1, 0],
                }
            ),
            origin="origin",
            development="development",
            columns=["amount"],
            cumulative=True,
        )
        .val_to_dev()
        .iloc[..., 0, -1]
        .sum()
        == 1
    )


def test_to_frame_on_zero(clrd):
    assert len((clrd * 0).latest_diagonal.to_frame(origin_as_datetime=False)) == 0


def test_valuation_vector():
    df = pd.DataFrame(
        {
            "Accident Date": [201508, 201608, 201708],
            "Valuation Date": [202111, 202111, 202111],
            "Loss": [110, 594, 696],
        }
    )

    tri = cl.Triangle(
        df,
        origin="Accident Date",
        development="Valuation Date",
        columns="Loss",
        cumulative=True,
        trailing=True,
    )

    assert int(tri.valuation_date.strftime("%Y%m")) == 202111


def test_single_entry():
    # triangle with one entry
    data = pd.DataFrame(
        {"origin": 2014, "valuation_date": "01.01.2017", "amount": 100}, index=[1]
    )
    cl_tri = cl.Triangle(
        data,
        origin="origin",
        development="valuation_date",
        columns="amount",
        cumulative=True,
    )

    # create a development constant
    dev_periods = cl_tri.val_to_dev().development.to_list()
    kwargs = {"patterns": {k: 1.5 for k in dev_periods}, "style": "ldf_"}
    cl_dev_constant = cl.DevelopmentConstant(**kwargs)

    # fit - this now works
    cl_dev_constant_fit = cl_dev_constant.fit(cl_tri.val_to_dev())

    # aim
    cl.Chainladder().fit(cl_dev_constant_fit.transform(cl_tri)).ultimate_


def test_origin_as_datetime_arg(clrd):
    from pandas.api.types import is_datetime64_any_dtype

    assert is_datetime64_any_dtype(clrd.to_frame(origin_as_datetime=True)["origin"])
    assert not is_datetime64_any_dtype(
        clrd.to_frame(origin_as_datetime=False)["origin"]
    )


def test_full_triangle_and_full_expectation(raa):
    raa_cum = raa
    assert raa_cum.is_cumulative == True

    raa_incr = raa_cum.cum_to_incr()
    assert raa_incr.is_cumulative == False
    assert raa_incr.incr_to_cum().is_cumulative == True
    assert raa_incr.incr_to_cum() == raa_cum

    cl_fit_incr = cl.Chainladder().fit(X=raa_incr)
    assert cl_fit_incr.X_.is_cumulative == False

    cl_fit_cum = cl.Chainladder().fit(X=raa_cum)
    assert cl_fit_cum.X_.is_cumulative == True

    assert cl_fit_incr.cdf_ == cl_fit_cum.cdf_
    assert cl_fit_incr.ultimate_ == cl_fit_cum.ultimate_

    assert (
        cl_fit_cum.full_expectation_ - cl_fit_incr.full_expectation_.incr_to_cum()
        < 0.00001
    )
    assert (
        cl_fit_cum.full_triangle_ - cl_fit_incr.full_triangle_.incr_to_cum() < 0.00001
    )
    assert (cl_fit_cum.full_triangle_ - raa_cum) - (
        cl_fit_incr.full_triangle_.incr_to_cum() - raa_incr.incr_to_cum()
    ) < 0.00001

    bf_fit_incr = cl.BornhuetterFerguson(apriori=1).fit(
        X=raa_incr, sample_weight=raa_incr.incr_to_cum().latest_diagonal * 0
    )
    assert bf_fit_incr.X_.is_cumulative == False

    bf_fit_cum = cl.BornhuetterFerguson(apriori=1).fit(
        X=raa_cum, sample_weight=raa_cum.latest_diagonal * 0
    )
    assert bf_fit_cum.X_.is_cumulative == True

    assert bf_fit_incr.cdf_ == bf_fit_cum.cdf_
    assert bf_fit_incr.ultimate_ == bf_fit_cum.ultimate_

    assert (
        bf_fit_cum.full_expectation_ - bf_fit_incr.full_expectation_.incr_to_cum()
        < 0.00001
    )
    assert (
        bf_fit_cum.full_triangle_ - bf_fit_incr.full_triangle_.incr_to_cum() < 0.00001
    )
    assert (bf_fit_cum.full_triangle_ - raa_cum) - (
        bf_fit_incr.full_triangle_.incr_to_cum() - raa_incr.incr_to_cum()
    ) < 0.00001

    assert (
        cl.Chainladder().fit(raa_incr).full_triangle_
        - cl.Chainladder().fit(raa_cum).full_triangle_.cum_to_incr()
        <= 0.0001
    )
    bk_fit_incr = cl.Benktander(apriori=1.00, n_iters=2).fit(
        X=raa_incr, sample_weight=raa_incr.incr_to_cum().latest_diagonal * 0
    )
    bk_fit_cum = cl.Benktander(apriori=1.00, n_iters=2).fit(
        X=raa_cum, sample_weight=raa_cum.latest_diagonal * 0
    )

    assert (
        bk_fit_cum.full_expectation_ - bk_fit_incr.full_expectation_.incr_to_cum()
        < 0.00001
    )
    assert (
        bk_fit_cum.full_triangle_ - bk_fit_incr.full_triangle_.incr_to_cum() < 0.00001
    )
    assert (bk_fit_cum.full_triangle_ - raa_cum) - (
        bk_fit_incr.full_triangle_.incr_to_cum() - raa_incr.incr_to_cum()
    ) < 0.00001


def test_halfyear_grain():
    data = pd.DataFrame(
        {"AccMo": [201409, 201503, 201603], "ValMo": [202203] * 3, "value": [100] * 3}
    )
    assert cl.Triangle(
        data=data, origin="AccMo", development="ValMo", columns="value", cumulative=True
    ).shape == (1, 1, 16, 1)


def test_predict(raa):
    raa_cum = raa
    assert cl.Chainladder().fit(raa_cum).X_.is_cumulative == True
    assert (
        cl.BornhuetterFerguson()
        .fit(raa_cum, sample_weight=raa_cum.latest_diagonal * 0 + 40000)
        .X_.is_cumulative
        == True
    )
    assert cl.Chainladder().fit_predict(raa_cum).is_cumulative == True
    assert (
        cl.BornhuetterFerguson()
        .fit_predict(raa_cum, sample_weight=raa_cum.latest_diagonal * 0 + 40000)
        .is_cumulative
        == True
    )

    raa_incr = raa.cum_to_incr()
    assert cl.Chainladder().fit(raa_incr).X_.is_cumulative == False
    assert (
        cl.BornhuetterFerguson()
        .fit(raa_incr, sample_weight=raa_incr.latest_diagonal * 0 + 40000)
        .X_.is_cumulative
        == False
    )
    assert cl.Chainladder().fit_predict(raa_incr).is_cumulative == False
    assert (
        cl.BornhuetterFerguson()
        .fit_predict(raa_incr, sample_weight=raa_incr.latest_diagonal * 0 + 40000)
        .is_cumulative
        == False
    )


def test_halfyear_development():
    df_sub = pd.read_csv(
        io.StringIO(
            """
        2011-01-01, 2011-01-01, 179.74
        2011-01-01, 2011-07-01, 664.94
        2011-01-01, 2012-01-01, 7471.75
        2011-01-01, 2012-07-01, 820.99
        2011-01-01, 2013-01-01, 908.77
        """
        ),
        names=["origin", "development", "paid"],
        parse_dates=["origin", "development"],
    )
    assert (
        type(
            cl.Triangle(
                data=df_sub,
                origin="origin",
                origin_format="%Y-%m-%d",
                development="development",
                development_format="%Y-%m-%d",
                columns="paid",
                cumulative=True,
            )
        )
        == cl.Triangle
    )

    data = [
        ["2010-01-01", "2011-06-30", "premium", 100.0],
        ["2010-01-01", "2011-12-31", "incurred", 100.0],
        ["2010-01-01", "2012-06-30", "premium", 200.0],
        ["2010-01-01", "2012-12-31", "incurred", 100.0],
        ["2010-01-01", "2013-12-31", "incurred", 200.0],
        ["2011-01-01", "2011-06-30", "premium", 100.0],
        ["2011-01-01", "2012-06-30", "premium", 200.0],
        ["2011-01-01", "2012-12-31", "incurred", 100.0],
        ["2011-01-01", "2013-12-31", "incurred", 200.0],
        ["2012-01-01", "2012-06-30", "premium", 200.0],
        ["2012-01-01", "2013-12-31", "incurred", 200.0],
    ]

    
    assert (
        type(
            cl.Triangle(
                data=pd.DataFrame(data, columns=["origin", "val_date", "idx", "value"]),
                index="idx",
                columns="value",
                origin="origin",
                development="val_date",
                cumulative=True,
            )
        )
    ) == cl.Triangle


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.methods import Benktander


class BornhuetterFerguson(Benktander):
    """The deterministic Bornhuetter Ferguson IBNR model

    Parameters
    ----------
    apriori: float, optional (default=1.0)
        Multiplier for the sample_weight used in the Bornhuetter Ferguson
        method. If sample_weight is already an apriori measure of ultimate,
        then use 1.0
    apriori_sigma: float, optional (default=0.0)
        Standard deviation of the apriori.  When used in conjunction with the
        bootstrap model, the model samples aprioris from a lognormal distribution
        using this argument as a standard deviation.
    random_state: int, RandomState instance or None, optional (default=None)
        Seed for sampling from the apriori distribution.  This is ignored when
        using as a deterministic method.
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random.

    Attributes
    ----------
    ultimate_: Triangle
        The ultimate losses per the method
    ibnr_: Triangle
        The IBNR per the method
    """

    def __init__(self, apriori=1.0, apriori_sigma=0.0, random_state=None):
        self.apriori = apriori
        self.apriori_sigma = apriori_sigma
        self.random_state = random_state

    def fit(self, X, y=None, sample_weight=None):
        """Applies the Bornhuetter-Ferguson technique to triangle **X**

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.
        y : None
            Ignored
        sample_weight : Triangle
            Required exposure to be used in the calculation.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.n_iters = 1
        super().fit(X, y, sample_weight)
        return self

    def predict(self, X, sample_weight=None):
        """Predicts the Bornhuetter-Ferguson ultimate on a new triangle **X**

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.
        sample_weight : Triangle
            Required exposure to be used in the calculation.

        Returns
        -------
        X_new: Triangle
            Loss data with Bornhuetter-Ferguson ultimate applied
        """
        return super().predict(X, sample_weight)



from chainladder.methods.base import MethodBase  # noqa (API import)
from chainladder.methods.chainladder import Chainladder  # noqa (API import)
from chainladder.methods.mack import MackChainladder  # noqa (API import)
from chainladder.methods.benktander import Benktander  # noqa (API import)
from chainladder.methods.bornferg import BornhuetterFerguson  # noqa (API import)
from chainladder.methods.capecod import CapeCod  # noqa (API import)
from chainladder.methods.expectedloss import ExpectedLoss  # noqa (API import)



import pytest
import chainladder as cl


def test_heatmap_render(raa):
    """The heatmap method should render correctly given the sample."""
    try:
        raa.heatmap()

    except:
        assert False


# def test_empty_triangle():
#     assert cl.Triangle()


def test_to_frame(raa):
    try:
        cl.Chainladder().fit(raa).cdf_.to_frame()
        cl.Chainladder().fit(raa).cdf_.to_frame(origin_as_datetime=False)
        cl.Chainladder().fit(raa).cdf_.to_frame(origin_as_datetime=True)
        cl.Chainladder().fit(raa).ultimate_.to_frame()
        cl.Chainladder().fit(raa).ultimate_.to_frame(origin_as_datetime=False)
        cl.Chainladder().fit(raa).ultimate_.to_frame(origin_as_datetime=True)

    except:
        assert False


def test_labels(xyz):
    assert (
        xyz.valuation_date.strftime("%Y-%m-%d %H:%M:%S.%f")
        == "2008-12-31 23:59:59.999999"
    )
    assert xyz.origin_grain == "Y"
    assert xyz.development_grain == "Y"
    assert xyz.shape == (1, 5, 11, 11)
    assert xyz.index_label == ["Total"]
    assert xyz.columns_label == ["Incurred", "Paid", "Reported", "Closed", "Premium"]
    assert xyz.origin_label == ["AccidentYear"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pandas as pd

from scipy.special import comb

from scipy.stats import binom, norm, rankdata

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from chainladder.core.triangle import Triangle


class DevelopmentCorrelation:
    """
    Mack (1997) test for correlations between subsequent development
    factors. Results should be within confidence interval range
    otherwise too much correlation

    Parameters
    ----------
    triangle: Triangle
        Triangle on which to estimate correlation between subsequent development
        factors.
    p_critical: float (default=0.5)
        Value between 0 and 1 representing the confidence level for the test. A
        value of 0.5 implies a 50% confidence. The default value is based on the example
        provided in the Mack 97 paper, the selection of which is justified on the basis of the
        test being only an approximate measure of correlations and the desire to detect
        correlations already in a substantial part of the triangle.

    Attributes
    ----------
    t_critical: DataFrame
        Boolean value for whether correlation is too high based on ``p_critical``
        confidence level.
    t_expectation: DataFrame
        Values representing the Spearman rank correlation
    t_variance: float
        Variance measure of Spearman rank correlation
    confidence_interval: tuple
        Range within which ``t_expectation`` must fall for independence assumption
        to be significant.
    """

    def __init__(self, triangle, p_critical: float = 0.5):
        self.p_critical = p_critical

        # Check that critical value is a probability
        validate_critical(p_critical=p_critical)

        if triangle.array_backend != "numpy":
            triangle = triangle.set_backend("numpy")
        xp = triangle.get_array_module()

        m1 = triangle.link_ratio

        # Rank link ratios by development period, assigning a score of 1 for the lowest
        m1_val = xp.apply_along_axis(
            func1d=rankdata, axis=2, arr=m1.values, nan_policy="omit"
        ) * (m1.values * 0 + 1)

        # Remove the last element from each column, and then rank again
        m2 = triangle[triangle.valuation < triangle.valuation_date].link_ratio
        m2.values = xp.apply_along_axis(
            func1d=rankdata, axis=2, arr=m2.values, nan_policy="omit"
        ) * (m2.values * 0 + 1)

        m1 = m2.copy()

        # remove the first column from m1 since it is not used in the comparison to m2
        m1.values = m1_val[..., : m2.shape[2], 1:]

        # Apply Spearman Rank Correlation formula
        # numerator is the one in formula G4 of the Mack 97 paper
        numerator = ((m1 - m2) ** 2).sum("origin")

        # remove last column because it was not part of the comparison with m2
        numerator.values = numerator.values[..., :-1]
        numerator.ddims = numerator.ddims[:-1]

        # I is the number of development periods in the triangle
        I = len(triangle.development)

        # k values are the column indexes for which we are calculating T_k
        k = xp.array(range(2, 2 + numerator.shape[3]))

        # denominator is the one in formula G4 of the Mack 97 paper
        denominator = ((I - k) ** 3 - I + k)[None, None, None]

        # complete formula G4, results in array of each T_k value
        self.t = 1 - 6 * xp.nan_to_num(numerator.values) / denominator

        # per Mack, weight is one less than the number of pairs for each T_k
        weight = (I - k - 1)[None, None, None]

        # Calculate big T, the weighted average of the T_k values
        t_expectation = (
            xp.sum(xp.nan_to_num(weight * self.t), axis=3) / xp.sum(weight, axis=3)
        )[..., None]

        idx = triangle.index.set_index(triangle.key_labels).index

        # variance is result of formula G6
        self.t_variance = 2 / ((I - 2) * (I - 3))

        # array of t values
        self.t = pd.DataFrame(self.t[0, 0, ...], columns=k, index=["T_k"])

        # array of weights
        self.weights = pd.DataFrame(weight[0, 0, ...], columns=k, index=["I-k-1"])

        # final big T
        self.t_expectation = pd.DataFrame(
            t_expectation[..., 0, 0], columns=triangle.vdims, index=idx
        )

        # table of Spearman's rank coefficients Tk, can be used to verify consistency with paper
        self.corr = pd.concat([self.t, self.weights])

        self.corr.columns.names = ["k"]

        # construct confidence interval based on selection of p_critical
        self.confidence_interval = (
            norm.ppf(0.5 - (1 - p_critical) / 2) * xp.sqrt(self.t_variance),
            norm.ppf(0.5 + (1 - p_critical) / 2) * xp.sqrt(self.t_variance),
        )

        # if T lies outside this range, we reject the null hypothesis
        self.t_critical = (self.t_expectation < self.confidence_interval[0]) | (
            self.t_expectation > self.confidence_interval[1]
        )

        # hypothesis test result, False means fail to reject the null hypothesis
        self.reject = self.t_critical.values[0][0]


class ValuationCorrelation:
    """
    Mack (1997) test for calendar year effect.A calendar period has impact
    across developments if the probability of the number of small (or large)
    development factors, Z, in that period occurring randomly is less than
    ``p_critical``

    Parameters
    ----------
    triangle: Triangle
        Triangle on which to test whether the calendar effects violate independence
        requirements of the chainladder method.
    p_critical: float (default=0.10)
        Value between 0 and 1 representing the confidence level for the test. 0.1
        implies 90% confidence.
    total: boolean
        Whether to calculate valuation correlation in total across all
        years (True) consistent with Mack 1993 or for each year separately
        (False) consistent with Mack 1997.

    Attributes
    ----------
    z : Triangle or DataFrame
        Z values for each Valuation Period
    z_critical : Triangle or DataFrame
        Boolean value for whether correlation is too high based on ``p_critical``
        confidence level.
    z_expectation : Triangle or DataFrame
        The expected value of Z.
    z_variance : Triangle or DataFrame
        The variance value of Z.
    """

    def __init__(self, triangle: Triangle, p_critical: float = 0.1, total: bool = True):

        def pZlower(z: int, n: int, p: float = 0.5) -> float:
            return min(1, 2 * binom.cdf(z, n, p))

        self.p_critical = p_critical

        # Check that critical value is a probability
        validate_critical(p_critical=p_critical)

        self.total = total
        triangle = triangle.set_backend("numpy")
        xp = triangle.get_array_module()
        lr = triangle.link_ratio

        # Rank link ratios for each column
        m1 = xp.apply_along_axis(
            func1d=rankdata, axis=2, arr=lr.values, nan_policy="omit"
        ) * (lr.values * 0 + 1)
        # print("med:\n", med)

        med = xp.nanmedian(a=m1, axis=2, keepdims=True)
        # print("med:\n", med)

        m1large = (xp.nan_to_num(m1) > med) + (lr.values * 0)
        m1small = (xp.nan_to_num(m1) < med) + (lr.values * 0)
        m2large = triangle.link_ratio
        m2large.values = m1large
        m2small = triangle.link_ratio
        m2small.values = m1small
        S = xp.nan_to_num(m2small.dev_to_val().sum(axis=2).set_backend("numpy").values)
        L = xp.nan_to_num(m2large.dev_to_val().sum(axis=2).set_backend("numpy").values)
        z = xp.minimum(L, S)
        n = L + S
        m = xp.floor((n - 1) / 2)
        c = comb(n - 1, m)
        EZ = (n / 2) - c * n / (2**n)
        VarZ = n * (n - 1) / 4 - c * n * (n - 1) / (2**n) + EZ - EZ**2
        if not self.total:
            T = []
            for i in range(0, xp.max(m1large.shape[2:]) + 1):
                T.append(
                    [
                        pZlower(i, j, 0.5)
                        for j in range(0, xp.max(m1large.shape[2:]) + 1)
                    ]
                )
            T = np.array(T)
            z_idx, n_idx = z.astype(int), n.astype(int)
            self.probs = T[z_idx, n_idx]
            z_critical = triangle[triangle.valuation > triangle.valuation.min()]
            # z_critical = z_critical[z_critical.development > z_critical.development.min()].dev_to_val().sum(
            #     "origin") * 0
            z_critical = z_critical.dev_to_val().dropna().sum("origin") * 0
            z_critical.values = np.array(self.probs) < p_critical
            z_critical.odims = triangle.odims[0:1]
            self.z_critical = z_critical
            self.z = self.z_critical.copy()
            self.z.values = z
            self.z_expectation = self.z_critical.copy()
            self.z_expectation.values = EZ
            self.z_variance = self.z_critical.copy()
            self.z_variance.values = VarZ
        else:
            ci2 = norm.ppf(0.5 - (1 - p_critical) / 2) * xp.sqrt(xp.sum(VarZ, axis=-1))
            self.range = (xp.sum(VarZ, axis=-1) + ci2, xp.sum(VarZ, axis=-1) - ci2)
            idx = triangle.index.set_index(triangle.key_labels).index
            self.z_critical = pd.DataFrame(
                (
                    (self.range[0] > VarZ.sum(axis=-1))
                    | (VarZ.sum(axis=-1) > self.range[1])
                )[..., 0],
                columns=triangle.vdims,
                index=idx,
            )
            self.z = pd.DataFrame(
                z.sum(axis=-1)[..., 0], columns=triangle.vdims, index=idx
            )
            self.z_expectation = pd.DataFrame(
                EZ.sum(axis=-1)[..., 0], columns=triangle.vdims, index=idx
            )
            self.z_variance = pd.DataFrame(
                VarZ.sum(axis=-1)[..., 0], columns=triangle.vdims, index=idx
            )


def validate_critical(p_critical: float) -> None:
    """
    Checks whether value passed to the p_critical parameter in ValuationCorrelation or DevelopmentCorrelation
    classes is a percentage, that is, between 0 and 1.

    Parameters
    ----------
    p_critical: float
        Critical value used to test null hypothesis in Mack correlation tests.
    """
    if 0 <= p_critical <= 1:
        pass
    else:
        raise ValueError("p_critical must be between 0 and 1.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
from scipy.optimize import least_squares
from chainladder.tails import TailBase
from chainladder.development import DevelopmentBase, Development


class TailBondy(TailBase):
    """Estimator for the Generalized Bondy tail factor.

    .. versionadded:: 0.6.0

    Parameters
    ----------
    earliest_age: int
        The earliest age from which the Bondy exponent is to be calculated.
        Defaults to latest age in the Triangle. Any available development
        age can be used.
    attachment_age: int (default=None)
        The age at which to attach the fitted curve.  If None, then the latest
        age is used. Measures of variability from original ``ldf_`` are retained
        when being used in conjunction with the MackChainladder method.

    Attributes
    ----------
    ldf_: Triangle
        ldf with tail applied.
    cdf_: Triangle
        cdf with tail applied.
    tail_: DataFrame
        Point estimate of tail at latest maturity available in the Triangle.
    b_: DataFrame
        The Bondy exponent
    earliest_ldf_: DataFrame
        The LDF associated with the ``earliest_age`` pick.
    projection_period: int
        The number of months beyond the latest available development age the
        `ldf_` and `cdf_` vectors should extend.

    See also
    --------
    TailCurve

    """

    def __init__(self, earliest_age=None, attachment_age=None, projection_period=12):
        self.earliest_age = earliest_age
        self.attachment_age = attachment_age
        self.projection_period = projection_period

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the tail will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        if self.attachment_age and self.attachment_age < self.earliest_age:
            raise ValueError("attachment_age must not be before earliest_age.")
        backend = X.array_backend
        if X.array_backend == "cupy":
            X = X.set_backend("numpy", deep=True)
        else:
            X = X.set_backend("numpy")
        xp = X.get_array_module()
        super().fit(X, y, sample_weight)

        if self.earliest_age is None:
            earliest_age = X.ddims[-2]
        else:
            earliest_age = X.ddims[
                int(
                    self.earliest_age / ({"Y": 12, "S": 6, "Q": 3, "M": 1}[X.development_grain])
                )
                - 1
            ]
        attachment_age = self.attachment_age if self.attachment_age else X.ddims[-2]
        obj = Development().fit_transform(X) if "ldf_" not in X else X
        b_optimized = []
        initial = xp.where(obj.ddims == earliest_age)[0][0] if earliest_age else 0
        for num in range(len(obj.vdims)):
            b0 = (xp.ones(obj.shape[0]) * 0.5)[:, None]
            data = xp.log(obj.ldf_.values[:, num, 0, initial:])
            b0 = xp.concatenate((b0, data[..., 0:1]), axis=1)
            b_optimized.append(
                least_squares(
                    TailBondy._solver, x0=b0.flatten(), kwargs={"data": data, "xp": xp}
                ).x
            )
        self.b_ = xp.concatenate(
            [item.reshape(-1, 2)[:, 0:1] for item in b_optimized], axis=1
        )[..., None, None]
        self.earliest_ldf_ = xp.exp(
            xp.concatenate(
                [item.reshape(-1, 2)[:, 1:2] for item in b_optimized], axis=1
            )[..., None, None]
        )
        if sum(X.ddims > earliest_age) > 1:
            tail = xp.exp(self.earliest_ldf_ * self.b_ ** (len(obj.ldf_.ddims) - 1))
        else:
            tail = self.ldf_.values[..., 0, initial]
        tail = tail ** (self.b_ / (1 - self.b_))
        f0 = self.ldf_.values[..., 0:1, initial : initial + 1]
        fitted = f0 ** (
            self.b_ ** (np.arange(sum(X.ddims >= earliest_age))[None, None, None, :])
        )
        fitted = xp.concatenate(
            (fitted, fitted[..., -1:] ** (self.b_ / (1 - self.b_))), axis=-1
        )
        fitted = xp.repeat(fitted, self.ldf_.shape[2], axis=2)
        rows = X.index.set_index(X.key_labels).index
        self.b_ = pd.DataFrame(self.b_[..., 0, 0], index=rows, columns=X.vdims)
        self.earliest_ldf_ = pd.DataFrame(
            self.earliest_ldf_[..., 0, 0], index=rows, columns=X.vdims
        )
        self.ldf_.values = xp.concatenate(
            (
                self.ldf_.values[..., : sum(X.ddims <= attachment_age)],
                fitted[..., -sum(X.ddims >= attachment_age) :],
            ),
            axis=-1,
        )
        self._get_tail_stats(obj)
        if backend == "cupy":
            self = self.set_backend(backend, inplace=True, deep=True)
        return self

    def transform(self, X):
        """Transform X.

        Parameters
        ----------
        X : Triangle
            Triangle must contain the ``ldf_`` development attribute.

        Returns
        -------
        X_new : Triangle
            New Triangle with tail factor applied to its development
            attributes.
        """
        X_new = super().transform(X)
        X_new.b_ = self.b_
        X_new.earliest_ldf_ = self.earliest_ldf_
        return X_new

    @staticmethod
    def _solver(b, data, xp):
        b = b.reshape(-1, 2)
        arange = xp.repeat(xp.arange(data.shape[-1])[None, :], data.shape[0], 0)
        out = data - (b[:, 1:2]) * b[:, 0:1] ** (arange)
        return out.flatten()



import pytest


def test_arithmetic_union(raa):
    assert raa.shape == (raa - raa[raa.valuation < "1987"]).shape
    assert raa[raa.valuation<'1986'] + raa[raa.valuation>='1986'] == raa


def test_arithmetic_across_keys(qtr):
    assert (qtr.sum(1) - qtr.iloc[:, 0]) == qtr.iloc[:, 1]


def test_arithmetic_1(raa):
    x = raa
    assert -(((x / x) + 0) * x) == -(+x)
    assert 1 - (x / x) ==  0 * x * 0


def test_rtruediv(raa):
    xp = raa.get_array_module()
    assert xp.nansum(abs(((1 / raa) * raa).values[0, 0] - raa.nan_triangle)) < 0.00001


def test_vector_division(raa):
    raa.latest_diagonal / raa


def test_multiindex_broadcast(clrd):
    clrd = clrd["CumPaidLoss"]
    clrd / clrd.groupby("LOB").sum()


def test_index_broadcasting(clrd):
    """ Basic broadcasting where b is a subset of a """
    assert ((clrd / clrd.sum()) - ((1 / clrd.sum()) * clrd)).sum().sum().sum() < 1e-4

def test_index_broadcasting2(clrd):
    """ b.key_labels are a subset of a.key_labels and b is missing some elements """
    a = clrd['CumPaidLoss']
    b = clrd['CumPaidLoss'].groupby('LOB').sum().iloc[:-1]
    c = a + b
    assert (a.index == c.index).all().all()


def test_index_broadcasting3(clrd):
    """ b.key_labels are a subset of a.key_labels and a is missing some elements """
    a = clrd[~clrd['LOB'].isin(['wkcomp', 'medmal'])]['CumPaidLoss']
    b = clrd['CumPaidLoss'].groupby('LOB').sum()
    c = a + b
    assert (a.index == c[~c['LOB'].isin(['wkcomp', 'medmal'])].index).all().all()
    assert len(c) - len(a) == 2


def test_index_broadcasting4(clrd):
    """ b should broadcast to a if b only has one index element """
    a = clrd['CumPaidLoss']
    b = clrd['CumPaidLoss'].groupby('LOB').sum().iloc[0]
    c = a + b
    assert (a.index == c.index).all().all()


@pytest.mark.xfail
def test_index_broadacsting4(clrd):
    """ If one triangle has key_labels that are not a subset of the other, then fail """
    a = clrd['CumPaidLoss']
    b = clrd['CumPaidLoss'].groupby('LOB').sum()
    idx = b.index
    idx['New Field'] = 'New'
    b.index = idx
    c = a + b

def test_index_broadcasting5(clrd):
    """ If a and b have shared key labels but no matching levels, then they will stack """
    a = clrd['CumPaidLoss'].iloc[:300]
    b = clrd['CumPaidLoss'].iloc[300:]
    c = a + b
    assert c.sort_index() == clrd['CumPaidLoss'].sort_index()


def test_index_broadacsting6(clrd):
    a = clrd['CumPaidLoss'].iloc[:100]
    b = clrd['CumPaidLoss'].iloc[50:150]
    c = clrd['CumPaidLoss'].iloc[50:100]
    d = a + b - c
    assert d.sort_index() == clrd['CumPaidLoss'].iloc[:150].sort_index()



import chainladder as cl
import numpy as np


def test_trend1(clrd):
    tri = clrd[["CumPaidLoss", "EarnedPremDIR"]].sum()
    lhs = (
        cl.CapeCod(0.05)
        .fit(tri["CumPaidLoss"], sample_weight=tri["EarnedPremDIR"].latest_diagonal)
        .ibnr_
    )
    rhs = (
        cl.CapeCod()
        .fit(
            cl.Trend(0.05).fit_transform(tri["CumPaidLoss"]),
            sample_weight=tri["EarnedPremDIR"].latest_diagonal,
        )
        .ibnr_
    )
    assert np.round(lhs, 0) == np.round(rhs, 0)


def test_trend2(raa):
    tri = raa
    assert (
        abs(
            cl.Trend(
                trends=[0.05, 0.05],
                dates=[(None, "1985"), ("1985", None)],
                axis="origin",
            )
            .fit(tri)
            .trend_
            * tri
            - tri.trend(0.05, axis="origin")
        )
        .sum()
        .sum()
        < 1e-6
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.core.io import EstimatorIO


class ParallelogramOLF(BaseEstimator, TransformerMixin, EstimatorIO):
    """
    Estimator to create and apply on-level factors to a Triangle object.  This
    is commonly used for premium vectors expressed as a Triangle object.

    Parameters
    ----------

    rate_history: pd.DataFrame
        A DataFrame with
    change_col: str
        The column containing the rate changes expressed as a decimal. For example,
        5% decrease should be stated as -0.05
    date_col: str
        A list-like set of effective dates corresponding to each of the changes
    vertical_line:
        Rates are typically stated on an effective date basis and premiums on
        and earned basis.  By default, this argument is False and produces
        parallelogram OLFs. If True, Parallelograms become squares.  This is
        commonly seen in Workers Compensation with benefit on-leveling or if
        the premium origin is also stated on an effective date basis.

    Attributes
    ----------

    olf_:
        A triangle representation of the on-level factors
    """

    def __init__(
        self,
        rate_history=None,
        change_col="",
        date_col="",
        approximation_grain="M",
        vertical_line=False,
    ):
        self.rate_history = rate_history
        self.change_col = change_col
        self.date_col = date_col
        self.approximation_grain = approximation_grain
        self.vertical_line = vertical_line

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Data to which the model will be applied.
        y: Ignored
        sample_weight: Ignored

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        from chainladder.utils.utility_functions import parallelogram_olf, concat

        if X.array_backend == "sparse":
            obj = X.set_backend("numpy")
        else:
            obj = X.copy()

        groups = list(set(X.key_labels).intersection(self.rate_history.columns))

        if len(groups) == 0:
            idx = obj
        else:
            idx = obj.groupby(groups).sum()

        kw = dict(
            start_date=X.origin[0].to_timestamp(how="s"),
            end_date=X.origin[-1].to_timestamp(how="e"),
            grain=X.origin_grain,
            vertical_line=self.vertical_line,
            approximation_grain=self.approximation_grain,
        )

        if len(groups) > 0:
            tris = []
            for item in idx.index.set_index(groups).iterrows():
                r = self.rate_history.set_index(groups).loc[item[0]].copy()
                r[self.change_col] = r[self.change_col] + 1
                r = (r.groupby(self.date_col)[self.change_col].prod() - 1).reset_index()
                date = r[self.date_col]
                values = r[self.change_col]
                olf = parallelogram_olf(values=values, date=date, **kw).values[
                    None, None
                ]
                if X.array_backend == "cupy":
                    olf = X.get_array_module().array(olf)
                tris.append((idx.loc[item[0]] * 0 + 1) * olf)
            self.olf_ = concat(tris, 0).latest_diagonal
        else:
            r = self.rate_history.copy()
            r[self.change_col] = r[self.change_col] + 1
            r = (r.groupby(self.date_col)[self.change_col].prod() - 1).reset_index()
            date = r[self.date_col]
            values = r[self.change_col]
            olf = parallelogram_olf(values=values, date=date, **kw)
            self.olf_ = ((idx * 0 + 1) * olf.values[None, None]).latest_diagonal
        return self

    def transform(self, X, y=None, sample_weight=None):
        """If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X: Triangle
            The triangle to be transformed

        Returns
        -------
            X_new: New triangle with transformed attributes.
        """
        X_new = X.copy()
        triangles = ["olf_"]
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new
